from nmdc_schema.migrators.adapters.adapter_base import AdapterBase
from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.6.0.part_1"
    _to_version = "11.6.0.part_2"

    def __init__(self, adapter: AdapterBase=None, logger=None):
        super().__init__(adapter, logger)

        self.ids_of_metatranscriptome_data_generation_with_no_output = set()
        self.ids_of_data_objects_to_update = set()
        self.data_object_type_update_map = {
            "Metagenome Raw Reads": "Metatranscriptome Raw Reads",
            "Metagenome Raw Read 1": "Metatranscriptome Raw Read 1",
            "Metagenome Raw Read 2": "Metatranscriptome Raw Read 2",
        }

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.ids_of_metatranscriptome_data_generation_with_no_output = set()
        self.ids_of_data_objects_to_update = set()

        self.adapter.do_for_each_document("data_generation_set", self.scan_data_generation_set)
        self.adapter.do_for_each_document("workflow_execution_set", self.scan_workflow_execution_set)
        self.logger.info(f"Found {len(self.ids_of_data_objects_to_update)} data objects to update.")

        self.adapter.process_each_document("data_object_set",[self.update_data_object_type])

    def scan_data_generation_set(self, data_generation: dict) -> None:
        r"""
        If the incoming data_generation object has analyte_category set to "metatranscriptome", and
        has_output is not set, add the id of the data_generation object to the set of
        ids_of_metatranscriptome_data_generation_with_no_output. This set will be used when scanning
        the workflow_execution_set. If the incoming data_generation object has analyte_category set
        to "metatranscriptome", and has_output is set, add each data_generation id in the has_output
        list to the set of ids_of_data_objects_to_update.

        >>> m = Migrator()
        >>> m.scan_data_generation_set({'id': 'nmdc:dgns-11-12345678', 'analyte_category': 'metatranscriptome', 'has_output': ["nmdc:dobj-11-abcd1234"]})
        >>> m.ids_of_data_objects_to_update
        {'nmdc:dobj-11-abcd1234'}
        >>> m.ids_of_metatranscriptome_data_generation_with_no_output
        set()
        >>> m = Migrator()
        >>> m.scan_data_generation_set({'id': 'nmdc:dgns-11-12345678', 'analyte_category': 'metagenome', 'has_output': ["nmdc:dobj-11-abcd1234"]})
        >>> m.ids_of_data_objects_to_update
        set()
        >>> m.ids_of_metatranscriptome_data_generation_with_no_output
        set()
        >>> m = Migrator()
        >>> m.scan_data_generation_set({'id': 'nmdc:dgns-11-12345678', 'analyte_category': 'metatranscriptome'})
        >>> m.ids_of_data_objects_to_update
        set()
        >>> m.ids_of_metatranscriptome_data_generation_with_no_output
        {'nmdc:dgns-11-12345678'}
        """
        if data_generation.get("analyte_category") == "metatranscriptome":
            if "has_output" not in data_generation:
                self.ids_of_metatranscriptome_data_generation_with_no_output.add(data_generation["id"])
            else:
                for output in data_generation.get("has_output", []):
                    self.ids_of_data_objects_to_update.add(output)

    def scan_workflow_execution_set(self, workflow_execution: dict) -> None:
        r"""
        If the incoming workflow_execution object has was_informed_by relationship to a
        data_generation object that was flagged in the scan_data_generation_set step (i.e. it is
        in the ids_of_metatranscriptome_data_generation_with_no_output set), and the
        workflow_execution object has a has_output field, add each data_generation id in the
        has_output list to the set of ids_of_data_objects_to_update.

        >>> m = Migrator()
        >>> m.ids_of_metatranscriptome_data_generation_with_no_output = {'nmdc:dgns-11-12345678'}
        >>> m.scan_workflow_execution_set({'id': 'nmdc:wfmsa-11-12345678', 'type': 'nmdc:MetagenomeSequencing', 'was_informed_by': 'nmdc:dgns-11-12345678', 'has_output': ["nmdc:dobj-11-abcd1234"]})
        >>> m.ids_of_data_objects_to_update
        {'nmdc:dobj-11-abcd1234'}
        >>> m = Migrator()
        >>> m.ids_of_metatranscriptome_data_generation_with_no_output = {'nmdc:dgns-11-12345678'}
        >>> m.scan_workflow_execution_set({'id': 'nmdc:wfmsa-11-12345678', 'type': 'nmdc:MetagenomeSequencing', 'was_informed_by': 'nmdc:dgns-11-98765432', 'has_output': ["nmdc:dobj-11-abcd1234"]})
        >>> m.ids_of_data_objects_to_update
        set()
        """
        if workflow_execution.get("type") != "nmdc:MetagenomeSequencing":
            return

        if "was_informed_by" in workflow_execution:
            data_generation_id = workflow_execution["was_informed_by"]
            if data_generation_id in self.ids_of_metatranscriptome_data_generation_with_no_output:
                for output in workflow_execution.get("has_output", []):
                    self.ids_of_data_objects_to_update.add(output)

    def update_data_object_type(self, data_object: dict) -> dict:
        r"""
        Update the data_object_type of data_objects whose ids are in the
        ids_of_data_objects_to_update set. If the current data_object_type is "Metagenome Raw
        Reads", change it to "Metatranscriptome Raw Reads". If the data_object_type is "Metagenome
        Raw Read 1", change it to "Metatranscriptome Raw Read 1". If the data_object_type is
        "Metagenome Raw Read 2", change it to "Metatranscriptome Raw Read 2". Otherwise, do nothing.

        >>> m = Migrator()

        # In reality, this would be set by scan_data_generation_set and scan_workflow_execution_set
        >>> m.ids_of_data_objects_to_update = {"nmdc:dobj-11-abcd1234"}

        >>> m.update_data_object_type({'id': 'nmdc:dobj-11-abcd1234', 'data_object_type': 'Metagenome Raw Reads'})
        {'id': 'nmdc:dobj-11-abcd1234', 'data_object_type': 'Metatranscriptome Raw Reads'}
        >>> m.update_data_object_type({'id': 'nmdc:dobj-11-abcd1234', 'data_object_type': 'Metagenome Raw Read 1'})
        {'id': 'nmdc:dobj-11-abcd1234', 'data_object_type': 'Metatranscriptome Raw Read 1'}
        >>> m.update_data_object_type({'id': 'nmdc:dobj-11-abcd1234', 'data_object_type': 'Metagenome Raw Read 2'})
        {'id': 'nmdc:dobj-11-abcd1234', 'data_object_type': 'Metatranscriptome Raw Read 2'}

        # Don't update data_objects that have a data_object_type that we don't care about
        >>> m.update_data_object_type({'id': 'nmdc:dobj-11-abcd1234', 'data_object_type': 'Metagenome Bins Barplot'})
        {'id': 'nmdc:dobj-11-abcd1234', 'data_object_type': 'Metagenome Bins Barplot'}

        # Don't update data_objects that are not in the ids_of_data_objects_to_update set
        >>> m.update_data_object_type({'id': 'nmdc:dobj-11-wxyz9871', 'data_object_type': 'Metagenome Raw Reads'})
        {'id': 'nmdc:dobj-11-wxyz9871', 'data_object_type': 'Metagenome Raw Reads'}
        """
        data_object_id = data_object["id"]
        if data_object_id not in self.ids_of_data_objects_to_update:
            return data_object

        data_object_type = data_object["data_object_type"]
        if data_object_type not in self.data_object_type_update_map:
            return data_object

        data_object["data_object_type"] = self.data_object_type_update_map[data_object_type]
        return data_object



from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.helpers import load_yaml_asset


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "8.1"
    _to_version = "9.0"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document("study_set", [
            self.fix_award_dois,
            self.fix_pub_dois,
            self.fix_massive,
            self.fix_ess_dive,
            self.remove_doi_slots
        ])

    def process_doi(self, study: dict, doi_list: list, doi_category: str):
        r"""
        For each update descriptor in the `doi_list` list; if the specified study's `id` matches that descriptor's `id`,
        this function appends a DOI dictionary to the study's `associated_dois` list (creating the list if necessary).

        >>> m = Migrator()
        >>> m.process_doi({'id': 123}, [{'id': 123, 'doi': 'a', 'doi_prov': 'b'}], 'c')  # `id` matches
        {'id': 123, 'associated_dois': [{'doi_value': 'a', 'doi_category': 'c', 'doi_provider': 'b'}]}
        >>> m.process_doi({'id': 123}, [{'id': 555, 'doi': 'a', 'doi_prov': 'b'}], 'c')  # `id` does not match
        {'id': 123}
        >>> m.process_doi(
        ...     {'id': 123, 'associated_dois': [{'doi_value': 'x', 'doi_category': 'y', 'doi_provider': 'z'}]},
        ...     [{'id': 123, 'doi': 'a', 'doi_prov': 'b'}], 'c'
        ... )  # study already has an `associated_dois` field
        {'id': 123, 'associated_dois': [{'doi_value': 'x', 'doi_category': 'y', 'doi_provider': 'z'}, {'doi_value': 'a', 'doi_category': 'c', 'doi_provider': 'b'}]}
        """

        id_value = study['id']
        for doi_updates in doi_list:
            if id_value == doi_updates['id']:
                new_doi = {
                    'doi_value': doi_updates['doi'],
                    'doi_category': doi_category,
                    'doi_provider': doi_updates['doi_prov']
                }
                study.setdefault('associated_dois', []).append(new_doi)

        return study

    def fix_award_dois(self, study: dict):
        """Moves current DOIs in the award_dois slot to the new associated_dois slot.
         It also changes some DOIs that were incorrectly labeled as award to dataset DOIs.
         It also adds three new award DOIs from JGI"""

        study_doi_data = load_yaml_asset('migrator_from_8_1_to_9_0/study_dois_changes.yaml')

        self.process_doi(
            study, study_doi_data['award_move_to_data'], 'dataset_doi')
        self.process_doi(study, study_doi_data['new_award_dois'], 'award_doi')
        self.process_doi(study, study_doi_data['award_doi_prov'], 'award_doi')

        return study

    def fix_pub_dois(self, study: dict):
        r"""
        Copies `publication_dois` values into a new slot named `associated_dois`.

        >>> m = Migrator()
        >>> m.fix_pub_dois({'id': 1, 'publication_dois': ['a']})  # test: a single DOI
        {'id': 1, 'publication_dois': ['a'], 'associated_dois': [{'doi_value': 'a', 'doi_category': 'publication_doi'}]}
        """

        study.setdefault('associated_dois', []).extend(
            {'doi_value': pub_doi, 'doi_category': 'publication_doi'}
            for pub_doi in study.get('publication_dois', [])
        )

        return study

    def fix_massive(self, study: dict):
        r"""
        Changes the one `massive_study_identifiers` value into a DOI and moves it to a new `associated_dois` slot.

        >>> m = Migrator()
        >>> m.fix_massive({'id': 123})
        {'id': 123, 'associated_dois': []}
        >>> m.fix_massive({'id': 123, 'massive_study_identifiers': ['not-the-one']})
        {'id': 123, 'massive_study_identifiers': ['not-the-one'], 'associated_dois': []}
        >>> m.fix_massive({'id': 123, 'massive_study_identifiers': ['MASSIVE:MSV000090886']})  # this is the one!
        {'id': 123, 'associated_dois': [{'doi_value': 'doi:10.25345/C58K7520G', 'doi_category': 'dataset_doi', 'doi_provider': 'massive'}]}
        """

        mass_id = 'MASSIVE:MSV000090886'
        mass_doi = 'doi:10.25345/C58K7520G'
        study.setdefault('associated_dois', []).extend(
            {'doi_value': mass_doi, 'doi_category': 'dataset_doi', 'doi_provider': 'massive'}
            for id in study.get('massive_study_identifiers', []) if id == mass_id)

        # remove the massive_study_identifiers slot if the id matches the one to be removed in associated_dois slot
        for doi_group in study['associated_dois']:
            if doi_group['doi_value'] == mass_doi:
                study.pop('massive_study_identifiers', None)

        return study

    def fix_ess_dive(self, study: dict):
        r"""
        Copies `ess_dive_datasets` values into a new slot named `associated_dois`.

        >>> m = Migrator()
        >>> m.fix_ess_dive({'id': 1, 'ess_dive_datasets': ['a']})  # test: a single DOI
        {'id': 1, 'ess_dive_datasets': ['a'], 'associated_dois': [{'doi_value': 'a', 'doi_category': 'dataset_doi', 'doi_provider': 'ess_dive'}]}
        """

        study.setdefault('associated_dois', []).extend(
            {'doi_value': dataset_doi, 'doi_category': 'dataset_doi',
             'doi_provider': 'ess_dive'}
            for dataset_doi in study.get('ess_dive_datasets', []))

        return study

    def remove_doi_slots(self, study: dict):
        r"""
        Removes slots that are no longer needed because their values have been moved to the `associated_dois` slot.

        >>> m = Migrator()
        >>> m.remove_doi_slots({'id': 123, 'associated_dois': []})  # empty `associated_dois` list gets removed
        {'id': 123}
        >>> m.remove_doi_slots({
        ...    'id': 123,
        ...    'associated_dois': [{'doi_value': 'a', 'doi_category': 'c', 'doi_provider': 'b'}],
        ...    'publication_dois': ['a'],
        ... })  # lists in which all values match an `associated_dois[].doi_value` get removed
        {'id': 123, 'associated_dois': [{'doi_value': 'a', 'doi_category': 'c', 'doi_provider': 'b'}]}
        """

        removal_slots = ['publication_dois', 'dataset_dois',
                         'award_dois', 'ess_dive_datasets', 'massive_study_identifiers']

        # Get dois from associated_dois
        associated_dois = [entry['doi_value']
                           for entry in study.get('associated_dois', [])]

        # Remove the old dois from the old doi slots
        for slot_name in removal_slots:
            study[slot_name] = [doi for doi in study.get(
                slot_name, []) if doi not in associated_dois]

            # Remove old doi slots if values are empty
            if not study[slot_name]:
                del study[slot_name]
            else:
                self.logger.error(f'Field {slot_name} of Study {study["id"]} was not falsy. Skipped slot deletion.')
                raise ValueError(f'Field {slot_name} of Study ({study["id"]}) is not empty.')

        # Remove associated_dois if empty (no dois were moved over and the slot is unnecessary)
        if not study['associated_dois']:
            del study['associated_dois']

        return study



from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.adapters.adapter_base import AdapterBase
from nmdc_schema.migrators.adapters.mongo_adapter import MongoAdapter
from nmdc_schema.migrators.migration_reporter import MigrationReporter
from typing import Optional

import logging

logging.basicConfig(level=logging.INFO, format='%(message)s')

class Migrator(MigratorBase):
    """
    Migrates data between two schemas.

    TUTORIAL: This is an example of a "migrator" class.
              It was designed for use during developer training and
              to serve as a template for production "migrator" classes.
    """

    """
    TUTORIAL: These two strings tell people the versions of the schema this
              "migrator" class was designed to migrate data between.
              
              -->  As part of creating a new "migrator" class, you will
                   populate its `_from_version` and `_to_version` strings.
    """
    _from_version = "1.0.0"  # https://github.com/microbiomedata/nmdc-schema/blob/1.0.0/
    _to_version = "EXAMPLE"  # in practice, this would be a real schema version; e.g., "7.8.0"

    def __init__(self, adapter: Optional[AdapterBase] = None, logger=None):
        """
        TUTORIAL: The __init__ function (constructor) initializes a new migrator instance.
                  This function runs automatically when you create a new migrator object.
                  
                  It accepts two optional parameters:
                  - adapter: The data store adapter (e.g., MongoAdapter, DictionaryAdapter)
                  - logger: A Python logger for outputting messages during migration
                  
                  If you don't provide these, the migrator will use default values.
                  
                  --> As part of creating a new migrator class, you typically won't need
                      to modify this function unless you need to initialize additional
                      instance variables specific to your migration logic.
        """
        super().__init__(adapter, logger)
        self.reporter = None

    def upgrade(self, commit_changes: bool = False):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        TUTORIAL: This is the `upgrade` function. You can think of it as the "main" function or
                  "entry point" of the migrator. Its job is to transform the database from
                  conforming to the original schema to conforming to the new schema.

                  The `upgrade` function accesses the database via an "adapter", instead of
                  accessing the database directly. That allows the same migrator to be used
                  with different types of data stores (e.g. MongoDB database, Python dictionary).

              --> As part of creating a new "migrator" class, you will implement an `upgrade` function.
              
        TUTORIAL: The `commit_changes` parameter controls whether changes are saved to the database:
                  - commit_changes=False (default): Changes are rolled back (dry run mode)
                  - commit_changes=True: Changes are committed and saved permanently
                  
                  This allows safe testing of migrations before applying them to production data.
              
        TUTORIAL: Example showing complete migration workflow with reporting:
        
        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> database = {
        ...     "study_set": [
        ...         {"id": "study1", "name": "Research Project"},
        ...         {"id": "study2"}  # study without name
        ...     ]
        ... }
        >>> m = Migrator(DictionaryAdapter(database))
        >>> result = m.upgrade()
        Processing collection: study_set
        >>> # Check transformation results
        >>> database["study_set"][0]["names"]
        ['Research Project']
        >>> "name" in database["study_set"][0]
        False
        >>> database["study_set"][1]["names"]
        []
        >>> # Check new collections were created
        >>> len(database["comment_set"])
        2
        >>> len(database["report_set"])
        2
        
        TUTORIAL: For MongoDB-based migrations, the commit_changes parameter controls transactions:
        
        Args:
            commit_changes (bool): If True, commits changes to database. If False (default), 
                                   rolls back changes for safe testing.
        """
        self.logger.setLevel(logging.INFO)
        # TUTORIAL: Initialize the migration reporter to track changes during migration
        self.reporter = MigrationReporter(self.logger)
        
        # TUTORIAL: Transaction handling for MongoDB migrations
        #           When using MongoDB, migrations run within a transaction for safety
        if isinstance(self.adapter, MongoAdapter):
            # MongoDB adapter - use a single transaction for all operations
            try:
                # TUTORIAL: Use the unified transaction method that gives full control
                self.adapter.execute_in_transaction(
                    operations_callback=self._perform_all_migration_operations,
                    commit_changes=commit_changes
                )
                
                if commit_changes:
                    self.logger.info("Transaction committed (changes have been saved)")
                else:
                    self.logger.info("Transaction rolled back (no changes were committed)")
                    
            except Exception as e:
                self.logger.error(f"Migration failed: {e}")
                raise
        else:
            # For non-MongoDB adapters (like DictionaryAdapter), just perform all operations
            # TUTORIAL: Dictionary adapter doesn't support transactions, so changes apply immediately
            self._perform_migration_operations()
            if not commit_changes:
                self.logger.info("Note: Dictionary adapter doesn't support rollback - changes are applied immediately")
        
        # TUTORIAL: Generate final migration report showing what was changed
        #           
        #           IMPORTANT LIMITATIONS: The migration reporter only tracks changes that you
        #           explicitly tell it about by calling its tracking methods. It does NOT
        #           automatically detect or report:
        #           
        #           - Collection creation/deletion (e.g., create_collection, drop_collection)
        #           - Document insertion/deletion (e.g., insert_document, delete_document)  
        #           - Schema changes (e.g., renaming fields, changing data types)
        #           - Index creation/modification
        #           - Database configuration changes
        #           - Any other structural changes to the database
        #           
        #           The reporter ONLY tracks what you manually report by calling methods like:
        #           - self.reporter.track_record_updated() for field value changes
        #           - self.reporter.track_record_processed() for processed records
        #           - self.reporter.track_missing_value() for missing data
        #           - self.reporter.track_unmapped_value() for unmappable data
        #           
        #           --> As part of implementing a migrator, you should call the appropriate
        #               reporter tracking methods throughout your transformation functions
        #               to ensure comprehensive reporting of the changes you make.
        #               
        #           --> The migration reporting system probably needs additional functions
        #               to track other types of changes (collections, indexes, etc.).
        if self.reporter:
            self.reporter.generate_final_report()
    
    def _perform_all_migration_operations(self, adapter, session):
        """
        TUTORIAL: This method performs ALL migration operations within a single transaction.
                  It receives the adapter and session from execute_in_transaction(), ensuring
                  proper session management and atomic operations.
                  
                  The callback signature (adapter, session) is required by execute_in_transaction().
                  While adapter is technically the same as self.adapter, both parameters are provided
                  for clean API design and to make the callback function self-contained.
                  
        Args:
            adapter: The MongoAdapter instance - provides access to database operations
                    (Note: This is the same instance as self.adapter, but passed for API clarity)
            session: The MongoDB session for transaction control - required for transactional operations
                    (Methods like insert_document, process_each_document need this for atomicity)
        """
        # TUTORIAL: Process existing documents (transactional)
        print("Processing collection: study_set")
        adapter.process_each_document("study_set", [self.allow_multiple_names], session)
        
        # TUTORIAL: Create collections (not transactional in MongoDB - happens immediately)
        adapter.create_collection("comment_set")
        adapter.create_collection("report_set")
        
        # TUTORIAL: Insert new documents (transactional)
        adapter.insert_document("comment_set", {"id": 1, "text": "Hello"}, session)
        adapter.insert_document("comment_set", {"id": 2, "text": "Goodbye"}, session)
        
        # TUTORIAL: Process documents to create reports (transactional)
        adapter.do_for_each_document("comment_set", self.create_report_based_upon_comment, session)


    def _perform_migration_operations(self):
        """
        TUTORIAL: Separated migration logic to support transaction handling.
                  This method contains all the actual data transformation operations.
        """
        # TUTORIAL: In this example, we will pass each document in the `study_set` collection through
        #           a processing pipeline that consists of a single function: `self.allow_multiple_names`.
        #
        print("Processing collection: study_set")
        self.adapter.process_each_document("study_set", [self.allow_multiple_names])

        # TUTORIAL: Create and populate a collection that doesn't exist in the `_from_version` schema.
        self.adapter.create_collection("comment_set")
        self.adapter.insert_document("comment_set", {"id": 1, "text": "Hello"})
        self.adapter.insert_document("comment_set", {"id": 2, "text": "Goodbye"})

        # TUTORIAL: Advanced: Create and populate another new collection; based upon the documents in a _different_ collection.
        self.adapter.create_collection("report_set")
        self.adapter.do_for_each_document("comment_set", self.create_report_based_upon_comment)

    def allow_multiple_names(self, study: dict) -> dict:
        """
        TUTORIAL: This is an example "transformation" function within the class.
                  Its job is to transform a dictionary that conforms to the
                  initial schema (in this case, version "1.0.0"), into one
                  that conforms to the target schema (in this case,
                  version "EXAMPLE"). That might involve adding a field,
                  converting a string into a list of strings, etc.

                  When this "transformation" function runs, it does three things:
                  1. It accepts a single dictionary as a parameter.
                  2. It transforms that dictionary.
                  3. It returns the transformed dictionary.

              --> As part of creating a new "migrator" class, you will
                  typically implement one or more "transformation" functions.

        TUTORIAL: The remaining part of this comment consists of several "doctests".
                  You can read more about doctests in the official Python documentation:
                  https://docs.python.org/3/library/doctest.html

                  Each doctest consists of (a) a snippet of Python code that calls this
                  function, followed on the next line by (b) whatever the author of the
                  test expects this function to return, if anything.

                  People can then read and run those doctests in order to learn about
                  what this function does / how this function behaves.

              --> As part of implementing a "transformation" function, you will
                  typically write a few doctests in the docstring of that function.

        >>> m = Migrator()  # creates a class instance on which we can call this function (i.e. this method)
        >>> m.allow_multiple_names({'id': 123})  # test: creates an empty `names` list
        {'id': 123, 'names': []}
        >>> m.allow_multiple_names({'id': 123, 'name': 'My project'})  # test: transfers existing name to `names` list
        {'id': 123, 'names': ['My project']}
        >>> m.allow_multiple_names({'id': 123, 'name': 'My project', 'foo': 'bar'})  # test: preserves other keys
        {'id': 123, 'foo': 'bar', 'names': ['My project']}
        >>> # Test edge cases
        >>> m.allow_multiple_names({'id': 456, 'name': ''})  # test: handles empty name
        {'id': 456, 'names': ['']}
        >>> m.allow_multiple_names({'id': 789, 'names': ['existing']})  # test: overwrites existing names
        {'id': 789, 'names': []}
        """

        # optional log message
        self.logger.info(f"Transforming study having id: {study['id']}")

        # Transform the dictionary.
        #
        # TUTORIAL: In this example, I am creating a new field named `names`, whose value is an empty list;
        #           then, if the original study has a `name` value, I'm storing that in the list and then
        #           deleting the `name` field.
        #
        study["names"] = []
        if "name" in study:
            original_name = study["name"]
            study["names"].append(original_name)
            del study["name"]
            
            # TUTORIAL: Track that we updated this record (converted single name to names list)
            if self.reporter:
                self.reporter.track_record_updated(
                    class_name="nmdc:Study",
                    slot_name="name", 
                    subclass_type="nmdc:Study",
                    source_value=original_name,
                    target_value=f"[{original_name}]"
                )
        else:
            # TUTORIAL: Track that we processed this record, but it had no name to convert
            if self.reporter:
                self.reporter.track_record_processed(
                    class_name="nmdc:Study",
                    slot_name="names",
                    subclass_type="nmdc:Study", 
                    value="[empty_list]"
                )

        # Return the transformed dictionary.
        return study

    def create_report_based_upon_comment(self, comment: dict) -> None:
        """
        Creates a report based upon the comment passed in.

        Note: Although this function will be passed a document from the `comment_set` collection,
              the function will actually modify a *different* collection instead.
              
        TUTORIAL: This demonstrates how to process documents from one collection and create
                  new documents in a different collection during migration.
                  
        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> database = {"comment_set": [], "report_set": []}
        >>> m = Migrator(DictionaryAdapter(database))
        >>> comment = {"id": 1, "text": "Hello world"}
        >>> m.create_report_based_upon_comment(comment)
        >>> len(database["report_set"])
        1
        >>> database["report_set"][0]["body"]
        'Someone wrote Hello world'
        >>> # Test with empty text
        >>> comment_empty = {"id": 2, "text": ""}
        >>> m.create_report_based_upon_comment(comment_empty)
        >>> database["report_set"][1]["body"]
        'Someone wrote '
        """
        report = {"body": f"Someone wrote {comment['text']}"}
        self.adapter.insert_document(collection_name="report_set", document=report)



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.9.0.part_1"
    _to_version = "11.9.0.part_2"

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        # Update each document in the `workflow_execution_set` collection so its `was_informed_by` value is a list.
        self.adapter.process_each_document(
            "workflow_execution_set", [self.update_was_informed_by]
        )

    def update_was_informed_by(self, workflow_execution: dict) -> dict:
        r"""
        Update the document's `was_informed_by` value to be a list (since the schema says the slot is now multivalued).

        >>> m = Migrator()
        >>> m.update_was_informed_by({'id': 1, 'was_informed_by': 'nmdc:dgns-11-000001'})
        {'id': 1, 'was_informed_by': ['nmdc:dgns-11-000001']}
        """
        self.logger.info(f"Processing WorkflowExecution: {workflow_execution['id']}")

        original_was_informed_by_value = workflow_execution["was_informed_by"]
        workflow_execution["was_informed_by"] = [original_was_informed_by_value]
        return workflow_execution



"""
Simplified migration reporting framework for NMDC schema migrators.

This module provides a streamlined reporting system that generates three specific tables:
1. Records updated table: class_processed, slot_processed, subclass_type, records_updated, source_value, target_value
2. Unmapped values table: class.slot, unmapped_value, total_count
3. Missing values table: class.slot, total_missing
"""

from collections import defaultdict
import logging
from typing import List, Optional

class MigrationReporter:
    """
    Simplified migration reporter that tracks and generates three specific summary tables.
    """
    
    def __init__(self, logger: logging.Logger):
        self.logger = logger
        
        # Data structures for the three tables
        # {(class, slot, subclass_type, source_value, target_value): {'conformant': count, 'non_conformant': count, 'updated': count}}
        self.records_updated = defaultdict(lambda: {'conformant': 0, 'non_conformant': 0, 'updated': 0})
        # {class_slot: {value: count}}
        self.unmapped_values = defaultdict(lambda: defaultdict(int))
        # {class_slot: count}
        self.missing_values = defaultdict(int)
    
    
    def track_record_updated(self, class_name: str, slot_name: str, subclass_type: str, 
                           source_value: str, target_value: str, count: int = 1) -> None:
        """Track a record update for the first table."""
        key = (class_name, slot_name, subclass_type, source_value, target_value)
        self.records_updated[key]['non_conformant'] += count
        self.records_updated[key]['updated'] += count
    
    def track_record_processed(self, class_name: str, slot_name: str, subclass_type: str, 
                             value: str, count: int = 1) -> None:
        """Track a record that was processed but didn't need updating (already conformant)."""
        key = (class_name, slot_name, subclass_type, value, value)
        self.records_updated[key]['conformant'] += count
    
    def track_unmapped_value(self, class_name: str, slot_name: str, value: str, count: int = 1) -> None:
        """Track an unmapped value that should cause validation failure."""
        class_slot = f"{class_name}.{slot_name}"
        self.unmapped_values[class_slot][value] += count
    
    def track_missing_value(self, class_name: str, slot_name: str, count: int = 1) -> None:
        """Track a missing value that should cause validation failure."""
        class_slot = f"{class_name}.{slot_name}"
        self.missing_values[class_slot] += count
    
    def generate_final_report(self) -> None:
        """Generate the final migration report with three tables."""
        self.logger.info("=" * 80)
        self.logger.info("MIGRATION SUMMARY")
        self.logger.info("=" * 80)
        
        self._generate_records_updated_table()
        self._generate_unmapped_values_table()
        self._generate_missing_values_table()
        
        self.logger.info("=" * 80)
    
    def _generate_records_updated_table(self) -> None:
        """Generate the records updated table."""
        self.logger.info("RECORDS UPDATED")
        self.logger.info("-" * 40)
        
        if not self.records_updated:
            self.logger.info("No records were updated.")
            self.logger.info("")
            return
        
        # Table header with source value before count columns
        header = (f"{'Class':<30} {'SubClassType':<25} {'Slot':<25} "
                 f"{'Source Value':<15} {'Conformant':<12} {'Not':<12} {'Updated':<8} {'Target Value':<15}")
        self.logger.info(header)
        self.logger.info(f"{'':<30} {'':<25} {'':<25} {'':<15} {'':<12} {'Conformant':<12} {'':<8} {'':<15}")
        self.logger.info("-" * 155)
        
        # Sort by class, then slot, then total records (descending)
        sorted_updates = sorted(self.records_updated.items(), 
                              key=lambda x: (x[0][0], x[0][1], -(x[1]['conformant'] + x[1]['non_conformant'])))
        
        for (class_name, slot_name, subclass_type, source_value, target_value), counts in sorted_updates:
            conformant_count = counts['conformant']
            non_conformant_count = counts['non_conformant']
            updated_count = counts['updated']
            
            # If subclass type is the same as class, leave it blank
            display_subclass = "" if subclass_type == class_name else subclass_type
            
            row = (f"{class_name:<30} {display_subclass:<25} {slot_name:<25} "
                   f"{source_value:<15} {conformant_count:<12} {non_conformant_count:<12} {updated_count:<8} {target_value:<15}")
            self.logger.info(row)
        
        self.logger.info("")
    
    def _generate_unmapped_values_table(self) -> None:
        """Generate the unmapped values table."""
        self.logger.info("UNMAPPED VALUES")
        self.logger.info("-" * 40)
        
        if not self.unmapped_values:
            self.logger.info("No unmapped values found.")
            self.logger.info("")
            return
        
        # Table header
        self.logger.info(f"{'Class.Slot':<35} {'Unmapped Value':<20} {'Total Count':<12}")
        self.logger.info("-" * 67)
        
        # Sort by class.slot, then by count (descending)
        for class_slot in sorted(self.unmapped_values.keys()):
            values = self.unmapped_values[class_slot]
            sorted_values = sorted(values.items(), key=lambda x: -x[1])
            
            for value, count in sorted_values:
                self.logger.info(f"{class_slot:<35} {value:<20} {count:<12}")
        
        self.logger.info("")
    
    def _generate_missing_values_table(self) -> None:
        """Generate the missing values table."""
        self.logger.info("MISSING VALUES")
        self.logger.info("-" * 40)
        
        if not self.missing_values:
            self.logger.info("No missing values found.")
            self.logger.info("")
            return
        
        # Table header
        self.logger.info(f"{'Class.Slot':<35} {'Total Missing':<12}")
        self.logger.info("-" * 47)
        
        # Sort by class.slot
        for class_slot in sorted(self.missing_values.keys()):
            count = self.missing_values[class_slot]
            self.logger.info(f"{class_slot:<35} {count:<12}")
        
        self.logger.info("")
    
    def validate_migration_success(self) -> None:
        """
        Validates that the migration was successful by checking for unmapped and missing values.
        Raises a comprehensive exception if any validation issues are found.
        
        Raises:
            ValueError: If unmapped values or missing values are found, indicating migration failure
        """
        errors = []
        
        # Check for unmapped values
        if self.unmapped_values:
            errors.append("UNMAPPED VALUES (units that could not be normalized):")
            for class_slot in sorted(self.unmapped_values.keys()):
                values = self.unmapped_values[class_slot]
                sorted_values = sorted(values.items(), key=lambda x: -x[1])
                for value, count in sorted_values:
                    errors.append(f"  {class_slot}: '{value}' ({count} occurrences)")
        
        # Check for missing values  
        if self.missing_values:
            errors.append("MISSING VALUES (required units that could not be inferred):")
            for class_slot in sorted(self.missing_values.keys()):
                count = self.missing_values[class_slot]
                errors.append(f"  {class_slot}: {count} missing units")
        
        # If any validation errors found, raise exception
        if errors:
            error_message = "Migration validation failed with the following issues:\n\n" + "\n".join(errors)
            error_message += "\n\nAll validation issues must be resolved before the migration can succeed."
            raise ValueError(error_message)


# Generic schema path utilities for migration reporting
def parse_schema_path(path: str) -> List[str]:
    """
    Parses a document path into schema-relevant components, filtering out array indices.

    Args:
        path: Path like "substances_used[0].volume" or "extraction.input_mass"

    Returns:
        List of schema slot names: ["substances_used", "volume"] or ["extraction", "input_mass"]

    >>> parse_schema_path("extraction.input_mass")  # no array
    ['extraction', 'input_mass']
    >>> parse_schema_path("substances_used[0].volume")  # 1-D array
    ['substances_used', 'volume']
    """
    if not path:
        return []

    parts = []
    for part in path.split('.'):
        if '[' in part:
            # Extract slot name, ignore array index
            slot_name = part.split('[')[0]
            if slot_name:  # Only add non-empty slot names
                parts.append(slot_name)
        else:
            parts.append(part)

    return parts


def get_clean_schema_path(path: str) -> str:
    """
    Converts a document path with array indices to a clean schema path for reporting.
    
    Args:
        path: Path like "substances_used[0].volume" or "extraction.input_mass"
        
    Returns:
        Clean schema path: "substances_used.volume" or "extraction.input_mass"

    >>> get_clean_schema_path("extraction.input_mass")  # no array
    'extraction.input_mass'
    >>> get_clean_schema_path("substances_used[0].volume")  # 1-D array
    'substances_used.volume'
    """
    if not path:
        return "root"
    
    # Parse and rejoin without array indices
    schema_parts = parse_schema_path(path)
    return '.'.join(schema_parts) if schema_parts else "root"


def resolve_class_from_schema_path(schema_view, root_class: str, slot_path: List[str]) -> Optional[str]:
    """
    Uses schema definitions to resolve the target class for a nested slot path.
    
    Args:
        schema_view: SchemaView instance
        root_class: Starting class name (without "nmdc:" prefix)
        slot_path: List of slot names leading to the target field
        
    Returns:
        str or None: The resolved class name (without nmdc: prefix), or None if not found
    """
    if not slot_path:
        return root_class
    
    current_class = root_class
    
    try:
        for slot_name in slot_path:
            # Get the slot definition for this class
            slot_def = schema_view.induced_slot(slot_name, current_class)
            if not slot_def or not slot_def.range:
                return None
            
            # Move to the range class
            current_class = slot_def.range
            
        return current_class
        
    except (AttributeError, KeyError, TypeError, ValueError):
        # If schema traversal fails due to invalid class/slot names or missing schema definitions
        return None


def get_most_specific_class_for_reporting(schema_view, document_root: dict, path: str) -> str:
    """
    Determines the most specific class type for reporting purposes.
    For nested objects, use schema resolution to find the immediate parent class.
    
    Args:
        schema_view: SchemaView instance  
        document_root: The root document for fallback context
        path: Path to a QuantityValue-ranged slot within the document
        
    Returns:
        str: The most specific class URI (e.g., "nmdc:PortionOfSubstance")
    """
    # Parse path to get components leading to the QuantityValue
    path_parts = parse_schema_path(path)
    if not path_parts:
        return document_root.get('type', 'unknown')
    
    # Remove the final field name to get the path to the containing object
    container_path = path_parts[:-1] if len(path_parts) > 1 else []
    
    # Start with document's root class
    doc_type = document_root.get('type', 'nmdc:Unknown')
    root_class = doc_type.replace('nmdc:', '') if doc_type.startswith('nmdc:') else doc_type
    
    # Use schema to resolve the class context for the container
    if container_path:
        target_class = resolve_class_from_schema_path(schema_view, root_class, container_path)
        if target_class:
            return f"nmdc:{target_class}"
    
    # Fallback to document type
    return doc_type





from nmdc_schema.migrators.migrator_base import MigratorBase

class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.8.0"
    _to_version = "11.9.0.part_1"

    wfe_mappings = {}
    data_objects_mapping = {}

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema.
        This migrator addresses issue https://github.com/microbiomedata/nmdc-schema/issues/1727.
        This will migrate fields from all `WorkflowExecution` records with type `nmdc:MetagenomeSequencing` to related `DataGeneration` records. Deletion of the `MetagenomeSequencing` records will occur outside of this migrator via the `/queries:run` API endpoint.
        
        The following fields will be migrated from the WorkflowExecution records to the DataGeneration records:
        - has_output
        
        Additionally, it will properly link the DataObjects to the associated DataGeneration records through the was_generated_by field.

        Important links to note:
        - The `was_informed_by` field in the WorkFlowExecution records will be used to link them to the DataGeneration records that were created to replace the deprecated WorkflowExecution records.
        - The `has_output` field in the WorkFlowExecution records will be used to identify the DataObjects that were generated by the records of interest, and will be added to the DataGeneration records.
        - the 'was_generated_by' field in the DataObjects will be updated to point to the corresponding DataGeneration records.
        """
        self.adapter.do_for_each_document("workflow_execution_set", self.store_we_ms_fields)
        self.adapter.process_each_document("data_object_set", [self.link_do_to_dg])
        self.adapter.process_each_document("data_generation_set", [self.migrate_fields_to_dg])


    def store_we_ms_fields(self, workflow_execution_record: dict) -> None:
        r"""
        For each workflow execution record that is of type MetagenomeSequencing, gather the following fields to fill the corresponding DataGeneration record:
        - has_output (used to fill the has_output field in the DataGeneration record, and used to link the DataObjects to the DataGeneration record)

        >>> m = Migrator()
        >>> m.store_we_ms_fields({"id": "nmdc:wfmsa-11-kd2tsp05.1", "type": "nmdc:Anything", "has_output": ["nmdc:dobj-11-nrjyjm34"], "was_informed_by": "nmdc:omprc-11-8yy07g22"})
        >>> m.wfe_mappings
        {}
        >>> m.data_objects_mapping
        {}
        >>> m.store_we_ms_fields({"id": "nmdc:wfmsa-11-kd2tsp04.1", "type": "nmdc:MetagenomeSequencing", "has_output": ["nmdc:dobj-11-nrjyjm33"], "was_informed_by": "nmdc:omprc-11-8yy07g21"})
        >>> m.wfe_mappings
        {'nmdc:omprc-11-8yy07g21': {'ms_id': 'nmdc:wfmsa-11-kd2tsp04.1', 'has_output': ['nmdc:dobj-11-nrjyjm33']}}
        >>> m.data_objects_mapping
        {'nmdc:dobj-11-nrjyjm33': 'nmdc:omprc-11-8yy07g21'}
        """
        if workflow_execution_record.get("type") == "nmdc:MetagenomeSequencing":
            ms_id = workflow_execution_record.get("id")
            dg_id = workflow_execution_record.get("was_informed_by")
            data_object_ids = workflow_execution_record.get("has_output", [])
            dg_record = {
                "ms_id": ms_id,
                "has_output": data_object_ids,
            }
            assert len(data_object_ids) == 1, "This migrator was implemented under the assumption that the `WorkflowExecution`'s `has_output` list contains exactly one item. A `has_output` list greater than 1 was found. Migrating this database will require modifying the migrator."
            self.data_objects_mapping[data_object_ids[0]] = workflow_execution_record.get("was_informed_by")
            # Add the dict to the wfe_mappings
            self.wfe_mappings[dg_id] = dg_record
            

    def link_do_to_dg(self, data_object_record: dict) -> dict:
        r"""
        For each data object, link it to the corresponding DataGeneration record through the was_generated_by field using the data_objects_mapping.
        
        >>> m = Migrator()
        >>> m.data_objects_mapping = {'nmdc:dobj-11-nrjyjm33': 'nmdc:omprc-11-8yy07g21'}
        >>> m.link_do_to_dg({"id": "nmdc:dobj-11-nrjyjm33", "was_generated_by": None})
        {'id': 'nmdc:dobj-11-nrjyjm33', 'was_generated_by': 'nmdc:omprc-11-8yy07g21'}
        >>> m.link_do_to_dg({"id": "nmdc:dobj-11-nrjyjm34", "was_generated_by": "nmdc:some_other_record"})
        {'id': 'nmdc:dobj-11-nrjyjm34', 'was_generated_by': 'nmdc:some_other_record'}
        """
        #create a list of the data object ids that we will be adding to
        data_object_ids = self.data_objects_mapping.keys()
        # get the id of the current data object
        do_id = data_object_record.get("id")
        # Check if the data object is an object of interest
        if do_id in data_object_ids:
            # Check if the data object already has a was_generated_by field that has values.
            if not data_object_record.get("was_generated_by"):
                data_object_record["was_generated_by"] = self.data_objects_mapping.get(do_id)
            else:
                # If it already has a was_generated_by field, we do not want to overwrite it
                self.logger.warning(f"Data object {do_id} already has a was_generated_by field, skipping update.")
        return data_object_record

    def migrate_fields_to_dg(self, data_generation_record: dict) -> dict:
        r"""
        Migrate the fields from the workflow execution record to the DataGeneration record.
        
        >>> m = Migrator()
        >>> m.wfe_mappings = {'nmdc:omprc-11-8yy07g21': {'ms_id': 'nmdc:wfmsa-11-kd2tsp04.1', 'has_output': ['nmdc:dobj-11-nrjyjm33']}}
        >>> m.migrate_fields_to_dg({"id": "nmdc:omprc-11-8yy07g21", "has_output": []})
        {'id': 'nmdc:omprc-11-8yy07g21', 'has_output': ['nmdc:dobj-11-nrjyjm33']}
        >>> m.migrate_fields_to_dg({"id": "nmdc:omprc-11-8yy07g22", "has_output": []})
        {'id': 'nmdc:omprc-11-8yy07g22', 'has_output': []}
        """
        # Get the target ids from the wfe_mappings
        target_ids = list(self.wfe_mappings.keys())
        dg_record_id = data_generation_record.get("id")

        # Check if the data generation record's id is one of the target ids
        if dg_record_id in target_ids:
            # get the corresponding mapping from wfe_mappings
            wfe_map = self.wfe_mappings[dg_record_id]
            dg_has_output = data_generation_record.pop("has_output", [])
            wfe_map_has_output = wfe_map.get("has_output", [])
            new_has_output = list(set(dg_has_output + wfe_map_has_output))
            # Migrate the has_output field
            data_generation_record["has_output"] = new_has_output
        return data_generation_record


from abc import ABC, abstractmethod
from typing import Dict, List
from logging import getLogger
from nmdc_schema.migrators.adapters.adapter_base import AdapterBase


class MigratorBase(ABC):
    """Base class containing properties and methods related to migrating data between two schema versions."""

    # The schema version from which this class migrates data.
    #
    # Note: This string is empty here. It will be populated within the migration-specific classes.
    #
    _from_version: str = ""

    # The schema version to which this class migrates data.
    #
    # Note: This string is empty here. It will be populated within the migration-specific classes.
    #
    _to_version: str = ""

    def __init__(self, adapter: AdapterBase = None, logger = None):
        """
        Initializes the migrator with an adapter and a logger.

        """
        # Store a reference to the specified adapter. Migrator methods can use it to manipulate the database.
        self.adapter = adapter

        # If a logger was specified, use it; otherwise, initialize one and use that.
        self.logger = getLogger(__name__) if logger is None else logger

        if self.adapter is None:
            self.logger.warning("No adapter was specified. Migration capability will be limited.")
    
    def _warn_if_commit_ignored(self, commit_changes: bool) -> None:
        """Warn if commit_changes=True but adapter doesn't support transactions.
        
        Examples:
        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> import logging
        >>> logging.basicConfig(level=logging.WARNING, format='%(levelname)s:%(message)s')
        >>> 
        >>> # Create a test migrator with dictionary adapter
        >>> class TestMigrator(MigratorBase):
        ...     _from_version = "test.1"
        ...     _to_version = "test.2"
        ...     def upgrade(self, commit_changes: bool = False) -> None:
        ...         self._warn_if_commit_ignored(commit_changes)
        >>> 
        >>> database = {"test_set": []}
        >>> adapter = DictionaryAdapter(database)
        >>> migrator = TestMigrator(adapter=adapter)
        >>> 
        >>> # Test with commit_changes=False (no warning)
        >>> migrator._warn_if_commit_ignored(False)
        >>> 
        >>> # Test with commit_changes=True (should warn) - output will be captured
        >>> migrator._warn_if_commit_ignored(True)  # doctest: +SKIP
        """
        if commit_changes:
            from nmdc_schema.migrators.adapters.mongo_adapter import MongoAdapter
            if not isinstance(self.adapter, MongoAdapter):
                self.logger.warning(
                    "commit_changes=True was specified, but the current adapter does not support transactions. "
                    "All changes will be applied immediately and cannot be rolled back. "
                    "Use MongoAdapter for transaction support."
                )

    @abstractmethod
    def upgrade(self, commit_changes: bool = False):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema.
        
        Args:
            commit_changes: If True, commits the changes. If False (default), performs a dry run or rollback.
        """
        pass

    @classmethod
    def get_origin_version(cls) -> str:
        """Returns the schema version this class migrates data from."""
        return cls._from_version

    @classmethod
    def get_destination_version(cls) -> str:
        """Returns the schema version this class migrates data to."""
        return cls._to_version




from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.5.1"
    _to_version = "11.6.0.part_1"

    def upgrade(self):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        Between schema version 11.5.1 and 11.6.0, the slot `sample_state_information` was removed from the schema class
        `PortionOfSubstance`. This function updates all instances of that schema class in the database accordingly.

        Note: In a database conforming to schema 11.5.1, instances of the `PortionOfSubstance` class cannot exist
              as documents in a collection. Instead, they can only exist as objects in the multivalued `substances_used`
              field of instances of specific schema classes. Those schema classes are:
              - [x] 1. `Extraction`
                - Docs: https://microbiomedata.github.io/nmdc-schema/Extraction/
                - All of these are documents in the `material_processing_set` collection,
                  having a type value of `nmdc:Extraction`.
              - [x] 2. `StorageProcess`
                - Docs: https://microbiomedata.github.io/nmdc-schema/StorageProcess/
                - All of these are documents in the `storage_process_set` collection,
                  having a type value of `nmdc:StorageProcess`.
              - [x] 3. `DissolvingProcess`
                - Docs: https://microbiomedata.github.io/nmdc-schema/DissolvingProcess/
                - All of these are documents in the `material_processing_set` collection,
                  having a type value of `nmdc:DissolvingProcess`.
              - [x] 4. `ChemicalConversionProcess`
                - Docs: https://microbiomedata.github.io/nmdc-schema/ChemicalConversionProcess/
                - All of these are documents in the `material_processing_set` collection,
                  having a type value of `nmdc:ChemicalConversionProcess`.
              - [x] 5. `MobilePhaseSegment`
                - Docs: https://microbiomedata.github.io/nmdc-schema/MobilePhaseSegment/
                - These are not represented by entire documents in any collection. Instead,
                  they are represented by objects having a type value of `nmdc:MobilePhaseSegment`
                  in the multivalued `ordered_mobile_phases` fields of instances of the following schema classes:
                  - [x] A. `ChromatographyConfiguration`
                    - Docs: https://microbiomedata.github.io/nmdc-schema/ChromatographyConfiguration/
                    - All of these are documents in the `configuration_set` collection,
                      having a type value of `nmdc:ChromatographyConfiguration`.
                  - [x] B. `ChromatographicSeparationProcess`
                    - Docs: https://microbiomedata.github.io/nmdc-schema/ChromatographicSeparationProcess/
                    - All of these are documents in the `material_processing_set` collection,
                      having a type value of `nmdc:ChromatographicSeparationProcess`.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> db = {
        ...     'material_processing_set': [
        ...         {
        ...             'id': 1,
        ...             'type': 'nmdc:Extraction',
        ...             'substances_used': [
        ...                 {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'solid', 'substance_role': 'acid'},
        ...             ]
        ...         },
        ...     ],
        ... }
        >>> a = DictionaryAdapter(database=db)
        >>> m = Migrator(adapter=a)
        >>> m.upgrade()
        >>> db['material_processing_set'][0]['substances_used'][0]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'acid'}
        """

        self.adapter.process_each_document("material_processing_set", [
            Migrator.remove_sample_state_information_field_within_material_processing_set_document,
        ])
        self.adapter.process_each_document("storage_process_set", [
            Migrator.remove_sample_state_information_field_within_storage_process_set_document,
        ])
        self.adapter.process_each_document("configuration_set", [
            Migrator.remove_sample_state_information_field_within_configuration_set_document,
        ])

    @staticmethod
    def remove_sample_state_information_field_within_material_processing_set_document(document: dict) -> dict:
        r"""
        If the document has a type value of `nmdc:Extraction`, `nmdc:DissolvingProcess`, `nmdc:ChemicalConversionProcess`,
        remove the `sample_state_information` field from any `PortionOfSubstance` instances in the document's `substances_used` field.
        
        If the document has a type value of `nmdc:ChromatographicSeparationProcess`, remove the `sample_state_information`
        field from any `MobilePhaseSegment` instances in the document's `ordered_mobile_phases` field.

        # Test: Document's type is `nmdc:Extraction` (i.e. one of the 3 types that undergo the same treatment).
        >>> transformed_document = Migrator.remove_sample_state_information_field_within_material_processing_set_document({
        ...     'id': 1,
        ...     'type': 'nmdc:Extraction',
        ...     'substances_used': [
        ...         {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'solid', 'substance_role': 'base'},
        ...         {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'gas', 'substance_role': 'acid'},
        ...     ]
        ... })
        >>> transformed_document["type"]
        'nmdc:Extraction'
        >>> transformed_document["substances_used"][0]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'base'}
        >>> transformed_document["substances_used"][1]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'acid'}

        # Test: Document's type is `nmdc:ChromatographicSeparationProcess`.
        >>> transformed_document = Migrator.remove_sample_state_information_field_within_material_processing_set_document({
        ...     'id': 1,
        ...     'type': 'nmdc:ChromatographicSeparationProcess',
        ...     'ordered_mobile_phases': [
        ...         {
        ...              'id': 4,
        ...              'type': 'nmdc:MobilePhaseSegment',
        ...              'substances_used': [
        ...                  {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'solid', 'substance_role': 'base'},
        ...                  {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'gas', 'substance_role': 'acid'},
        ...              ]
        ...         }
        ...     ],
        ... })
        >>> transformed_document["type"]
        'nmdc:ChromatographicSeparationProcess'
        >>> transformed_document["ordered_mobile_phases"][0]["substances_used"][0]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'base'}
        >>> transformed_document["ordered_mobile_phases"][0]["substances_used"][1]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'acid'}
        """

        if document.get("type") in ["nmdc:Extraction", "nmdc:DissolvingProcess", "nmdc:ChemicalConversionProcess"]:
            if "substances_used" in document:
                Migrator.remove_sample_state_information_from_substances_used(document.get("substances_used", []))

        elif document.get("type") == "nmdc:ChromatographicSeparationProcess":
            if "ordered_mobile_phases" in document:
                for item in document.get("ordered_mobile_phases", []):
                    if item.get("type") == "nmdc:MobilePhaseSegment":
                        if "substances_used" in item:
                            Migrator.remove_sample_state_information_from_substances_used(item.get("substances_used", []))
        
        return document

    @staticmethod
    def remove_sample_state_information_field_within_storage_process_set_document(document: dict) -> dict:
        r"""
        If the document has a type value of `nmdc:StorageProcess`, remove the `sample_state_information` field from each
        `PortionOfSubstance` instance in the document's `substances_used` field.

        # Test: Document's type is `nmdc:StorageProcess`.
        >>> transformed_document = Migrator.remove_sample_state_information_field_within_storage_process_set_document({
        ...     'id': 1,
        ...     'type': 'nmdc:StorageProcess',
        ...     'substances_used': [
        ...         {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'solid', 'substance_role': 'base'},
        ...         {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'gas', 'substance_role': 'acid'},
        ...     ]
        ... })
        >>> transformed_document["type"]
        'nmdc:StorageProcess'
        >>> transformed_document["substances_used"][0]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'base'}
        >>> transformed_document["substances_used"][1]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'acid'}
        """

        if document.get("type") == "nmdc:StorageProcess":
            if "substances_used" in document:
                Migrator.remove_sample_state_information_from_substances_used(document.get("substances_used", []))
        
        return document

    @staticmethod
    def remove_sample_state_information_field_within_configuration_set_document(document: dict) -> dict:
        r"""
        If the document has a type value of `nmdc:ChromatographyConfiguration`, remove the `sample_state_information`
        field from any `MobilePhaseSegment` instances in the document's `ordered_mobile_phases` field.

        # Test: Document's type is `nmdc:ChromatographyConfiguration`.
        >>> transformed_document = Migrator.remove_sample_state_information_field_within_configuration_set_document({
        ...     'id': 1,
        ...     'type': 'nmdc:ChromatographyConfiguration',
        ...     'ordered_mobile_phases': [
        ...         {
        ...              'id': 4,
        ...              'type': 'nmdc:MobilePhaseSegment',
        ...              'substances_used': [
        ...                  {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'solid', 'substance_role': 'base'},
        ...                  {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'gas', 'substance_role': 'acid'},
        ...              ]
        ...         }
        ...     ],
        ... })
        >>> transformed_document["type"]
        'nmdc:ChromatographyConfiguration'
        >>> transformed_document["ordered_mobile_phases"][0]["substances_used"][0]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'base'}
        >>> transformed_document["ordered_mobile_phases"][0]["substances_used"][1]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'acid'}
        """

        if document.get("type") == "nmdc:ChromatographyConfiguration":
            if "ordered_mobile_phases" in document:
                for item in document.get("ordered_mobile_phases", []):
                    if item.get("type") == "nmdc:MobilePhaseSegment":
                        if "substances_used" in item:
                            Migrator.remove_sample_state_information_from_substances_used(item.get("substances_used", []))
        
        return document

    @staticmethod
    def remove_sample_state_information_from_substances_used(substances_used: list) -> list:
        r"""
        Removes the `sample_state_information` key from each `nmdc:PortionOfSubstance` dictionary
        in the specified `substances_used` list. Although the function modifies the list _in place_,
        it also _returns_ the modified list to facilitate unit testing (see doctests below).

        # Test: Returns empty list as is.
        >>> Migrator.remove_sample_state_information_from_substances_used([])
        []

        # Test: Returns original list if no objects have the target `type`.
        >>> Migrator.remove_sample_state_information_from_substances_used([
        ...     {'type': 'Other', 'sample_state_information': 'solid', 'substance_role': 'base'},
        ...     {'type': 'Other', 'sample_state_information': 'gas', 'substance_role': 'acid'},
        ... ])
        [{'type': 'Other', 'sample_state_information': 'solid', 'substance_role': 'base'}, {'type': 'Other', 'sample_state_information': 'gas', 'substance_role': 'acid'}]

        # Test: Returns original list if no objects have the `sample_state_information` field.
        >>> Migrator.remove_sample_state_information_from_substances_used([
        ...     {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'acid'},
        ... ])
        [{'type': 'nmdc:PortionOfSubstance', 'substance_role': 'acid'}]

        # Test: Removes the `sample_state_information` from only—and all—objects having the target `type`.
        >>> transformed_list = Migrator.remove_sample_state_information_from_substances_used([
        ...     {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'solid', 'substance_role': 'base'},
        ...     {'type': 'Other', 'sample_state_information': 'solid', 'substance_role': 'base'},
        ...     {'type': 'nmdc:PortionOfSubstance', 'sample_state_information': 'gas', 'substance_role': 'acid'},
        ... ])
        >>> len(transformed_list)
        3
        >>> transformed_list[0]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'base'}
        >>> transformed_list[1]
        {'type': 'Other', 'sample_state_information': 'solid', 'substance_role': 'base'}
        >>> transformed_list[2]
        {'type': 'nmdc:PortionOfSubstance', 'substance_role': 'acid'}
        """

        for substance_used in substances_used:
            if substance_used.get("type") == "nmdc:PortionOfSubstance":
                substance_used.pop("sample_state_information", None)  # deletes the key if it is present

        return substances_used



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.11.0.part_1"
    _to_version = "11.11.0.part_2"

    def upgrade(self, commit_changes: bool = True) -> None:
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        """

        self.adapter.do_for_each_document(
            "data_object_set", self.confirm_permissible_values_are_absent
        )

    def confirm_permissible_values_are_absent(self, data_object: dict) -> None:
        r"""
        If the data object has the data_object_type of "Metagenome Bins" or "Centrifuge Classification Report" raise an exception.

        >>> m = Migrator()
 
        # Test: data_object_type of "Metagenome Bins"
        >>> m.confirm_permissible_values_are_absent({"id": 1, "type": "nmdc:DataObject", "data_object_type": "Metagenome Bins"})
        Traceback (most recent call last):
            ...
        ValueError: DataObject 1 has value: Metagenome Bins

        # Test: data_object_type of "Centrifuge Classification Report"
        >>> m.confirm_permissible_values_are_absent({"id": 2, "type": "nmdc:DataObject", "data_object_type": "Centrifuge Classification Report"})
        Traceback (most recent call last):
            ...
        ValueError: DataObject 2 has value: Centrifuge Classification Report

        # Test: valid data_object_type
        >>> m.confirm_permissible_values_are_absent({"id": 3, "type": "nmdc:DataObject", "data_object_type": "Virus Summary"}) is None
        True
        """

        data_object_type_value = data_object.get("data_object_type")
        data_object_id = data_object.get("id")
        if data_object_type_value in ["Metagenome Bins", "Centrifuge Classification Report"]:
            raise ValueError(
                 f"DataObject {data_object_id} has value: {data_object_type_value}"
            )




from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.3.0"
    _to_version = "11.4.0"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document("workflow_execution_set", [self.set_metab_analysis_category])

    def set_metab_analysis_category(self, workflow: dict) -> dict:
        r"""
        If the workflow execution record is of the type "nmdc:MetabolomicsAnalysis" and it has a `has_metabolite_identifications` field,
        add field `metabolomics_analysis_category` and assign it the value "gc_ms_metabolomics". If the record does not 
        have a `has_metabolite_identifications` field, it is assigned the value "lc_ms_lipidomics".

        >>> m = Migrator()
        >>> m.set_metab_analysis_category({'id': 123, 'type': 'nmdc:MetabolomicsAnalysis', 'has_metabolite_identifications': []})
        {'id': 123, 'type': 'nmdc:MetabolomicsAnalysis', 'has_metabolite_identifications': [], 'metabolomics_analysis_category': 'gc_ms_metabolomics'}
        >>> m.set_metab_analysis_category({'id': 123, 'type': 'nmdc:MetabolomicsAnalysis'})  # does not have has_metabolite_identifications field, therefore it's a lipid analysis
        {'id': 123, 'type': 'nmdc:MetabolomicsAnalysis', 'metabolomics_analysis_category': 'lc_ms_lipidomics'}
        >>> m.set_metab_analysis_category({'id': 123, 'type': 'nmdc:Metaproteomics'})  # not a metabolomics analysis
        {'id': 123, 'type': 'nmdc:Metaproteomics'}
        """

        if workflow["type"] == "nmdc:MetabolomicsAnalysis":
            if "has_metabolite_identifications" in workflow:
                workflow["metabolomics_analysis_category"] = "gc_ms_metabolomics"
            else:
                workflow["metabolomics_analysis_category"] = "lc_ms_lipidomics"
        return workflow


from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.partials.migrator_from_11_11_0_to_11_13_0 import (
    get_migrator_classes,
)


class Migrator(MigratorBase):
    r"""
    Migrates a database between two schemas.
    """

    _from_version = "11.11.0"
    _to_version = "11.13.0"

    def upgrade(self, commit_changes: bool = False) -> None:
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        This migrator uses partial migrators. It runs them in the order in which they are returned by
        the `get_migrator_classes` function.
        
        Args:
            commit_changes: If True, commits the changes. If False (default), performs a dry run or rollback.
        """

        migrator_classes = get_migrator_classes()
        num_migrators = len(migrator_classes)
        for idx, migrator_class in enumerate(migrator_classes):
            self.logger.info(f"Running migrator {idx + 1} of {num_migrators}")
            self.logger.debug(
                f"Migrating from {migrator_class.get_origin_version()} "
                f"to {migrator_class.get_destination_version()}"
            )
            migrator = migrator_class(adapter=self.adapter, logger=self.logger)
            migrator.upgrade(commit_changes=commit_changes)



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""
    Migrates a database between two schemas.

    This migrator removes the `has_peptide_quantificiations` field from all documents that represent an instance of the `MetaproteomicsAnalysis` class.
    """
    _from_version = "11.1.0.part_2"
    _to_version = "11.1.0.part_3"

    def upgrade(self) -> None:
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        """

        self.adapter.process_each_document(collection_name="workflow_execution_set", pipeline=[self.remove_has_peptide_quantificiations])

    def remove_has_peptide_quantificiations(self, workflow_doc: dict) -> dict:
        r"""
        Removes the `has_peptide_quantifications` field from all documents that represent an instance of the `MetaproteomicsAnalysis` class.

        Only documents that have a `type` field with the value `nmdc:MetaproteomicsAnalysis` are modified and they are the only documents that can have the `has_peptide_quantificiations` field.

        # Documents with the `has_peptide_quantificiations` field are modified
        >>> m = Migrator()
        >>> m.remove_has_peptide_quantificiations({'id': 'ID1', 'has_peptide_quantifications': 'a', 'type': 'nmdc:MetaproteomicsAnalysis'})
        {'id': 'ID1', 'type': 'nmdc:MetaproteomicsAnalysis'}
        >>> m.remove_has_peptide_quantificiations({'id': 'ID2', 'has_peptide_quantifications': 'b', 'foo': 'bar', 'type': 'nmdc:MetaproteomicsAnalysis'})
        {'id': 'ID2', 'foo': 'bar', 'type': 'nmdc:MetaproteomicsAnalysis'}

        # Non-MetaproteomicsAnalysis documents are not modified
        >>> m.remove_has_peptide_quantificiations({'id': 'ID3', 'type': 'nmdc:MetabolomicsAnalysis'})
        {'id': 'ID3', 'type': 'nmdc:MetabolomicsAnalysis'}

        """

        if workflow_doc.get("type") == "nmdc:MetaproteomicsAnalysis":
            if "has_peptide_quantifications" in workflow_doc.keys():
                del workflow_doc["has_peptide_quantifications"]

        return workflow_doc


from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "9.3"
    _to_version = "10.0"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document(
            "extraction_set", [self.move_extraction_qc_status]
        )

    def move_extraction_qc_status(self, extraction: dict) -> dict:
        r"""
        Updates the `Extraction` so that the value originally in its `quality_control_report.status` field, if any,
        is stored in a new field named `qc_status`; and the `quality_control_report` field no longer exists.

        >>> m = Migrator()
        >>> m.move_extraction_qc_status({'id': 123, 'quality_control_report': {'status': 'pass'}})
        {'id': 123, 'qc_status': 'pass'}
        >>> m.move_extraction_qc_status({'id': 123, 'quality_control_report': {'status': 'fail'}})
        {'id': 123, 'qc_status': 'fail'}
        >>> m.move_extraction_qc_status({'id': 123, 'quality_control_report': {'status': 'pass', 'name': 'n'}})
        {'id': 123, 'qc_status': 'pass'}
        >>> m.move_extraction_qc_status({'id': 123, 'quality_control_report': {'status': 'fail', 'name': 'n'}})
        {'id': 123, 'qc_status': 'fail'}
        >>> m.move_extraction_qc_status({'id': 123, 'quality_control_report': {'name': 'n'}})
        {'id': 123}
        >>> m.move_extraction_qc_status({'id': 123, 'quality_control_report': {}})
        {'id': 123}
        """
        self.logger.info(f"Migrating Extraction: {extraction['id']}")
        if "quality_control_report" in extraction:
            if "status" in extraction["quality_control_report"]:
                extraction["qc_status"] = extraction["quality_control_report"]["status"]

            # Delete the key, even if it contains sub-keys other than "status".
            del extraction["quality_control_report"]
        return extraction



from typing import Optional

from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.nmdc_data import get_nmdc_schema_definition
from linkml_runtime import SchemaView


def create_schema_view() -> SchemaView:
    """
    Returns a LinkML SchemaView instance that can be used to traverse the schema.
    
    >>> isinstance(create_schema_view(), SchemaView)
    True
    """
    schema_definition = get_nmdc_schema_definition()
    schema_view = SchemaView(schema_definition)
    return schema_view


class Migrator(MigratorBase):
    """Migrates data from schema X to PR10"""

    _from_version = "X"
    _to_version = "PR10"

    def upgrade(self):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        Note: We added a prefix, `SKIP: `, to the doctests in this method to prevent them from being run.
              Reason for skipping: The tests target a function that—through its use of a `SchemaView` that
              is bound to the schema currently in the repository—requires that specific schema elements
              be defined in the schema, and those elements are not defined in the schema anymore. Note that
              the prefix string is arbitrary (I chose `SKIP: ` in an attempt to make it self-documenting).

        >>> from sys import stderr
        >>> print("⚠️  Skipping doctests", file=stderr)

        SKIP: >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        SKIP: >>> database = {
        SKIP: ...     "data_object_set": [
        SKIP: ...         {'id': 1},
        SKIP: ...         {'id': 2, 'type': 'old'}
        SKIP: ...     ],
        SKIP: ...     "library_preparation_set": [
        SKIP: ...         {'id': 1},
        SKIP: ...         {'id': 2, 'type': 'old'},
        SKIP: ...         {'id': 3, 'protocol_link': {'name': 'nombre'}},
        SKIP: ...         {'id': 4, 'protocol_link': {'name': 'nombre', 'type': 'old'}}
        SKIP: ...     ],
        SKIP: ... }
        SKIP: >>> m = Migrator(adapter=DictionaryAdapter(database=database))
        SKIP: >>> m.upgrade()
        SKIP: >>> all(document['type'] == 'nmdc:DataObject' for document in database['data_object_set'])
        SKIP: True
        SKIP: >>> all(document['type'] == 'nmdc:LibraryPreparation' for document in database['library_preparation_set'])
        SKIP: True
        SKIP:
        SKIP: Confirm a `type` field has been added to the inline `protocol_link` instance that lacked one.
        SKIP: >>> library_prep_3 = next(document for document in database['library_preparation_set'] if document['id'] == 3)
        SKIP: >>> library_prep_3['protocol_link']['name']
        SKIP: 'nombre'
        SKIP: >>> library_prep_3['protocol_link']['type']
        SKIP: 'nmdc:Protocol'
        SKIP:
        SKIP: Confirm the `type` value has been updated on the inline `protocol_link` instance that had an incorrect one.
        SKIP: >>> library_prep_4 = next(document for document in database['library_preparation_set'] if document['id'] == 4)
        SKIP: >>> library_prep_4['protocol_link']['name']
        SKIP: 'nombre'
        SKIP: >>> library_prep_4['protocol_link']['type']
        SKIP: 'nmdc:Protocol'
        """

        # Get a dictionary of slots and the class uris of their range if they have inlined classes as their range.
        view = create_schema_view()

        slots_with_inlined_classes = {}
        classes_with_inlined_classes = ["Biosample", 
                                        "Study", 
                                        "Extraction", 
                                        "MetabolomicsAnalysis" , 
                                        "MetaproteomicsAnalysis", 
                                        "MagsAnalysis",
                                        "ReadQcAnalysis",
                                        "DataGeneration",
                                        "LibraryPreparation",
                                        "Pooling",
                                        "MetagenomeAnnotation",
                                        "MetagenomeAssembly",
                                        "MetagenomeSequencing",
                                        "NomAnalysis",
                                        "ReadBasedTaxonomyAnalysis",
                                        "MetatranscriptomeAnnotation",
                                        "MetatranscriptomeAssembly",
                                        "MetatranscriptomeExpressionAnalysis",
                                        "CollectingBiosamplesFromSite"
                                        ]
        for nmdc_class in classes_with_inlined_classes:
            induced_slots = view.class_induced_slots(nmdc_class)
            for slot_def in induced_slots:
                slot_name = slot_def.name
                slot_range = slot_def.range
                element = view.get_element(slot_range)
                metatype = type(element).class_name
                if metatype == 'class_definition':
                    class_identifying_slot = view.get_identifier_slot(slot_range)
                    if not class_identifying_slot:
                        # need to hardcode because schema prefix is not NMDC
                        if slot_name == "has_credit_associations":
                            class_uri = "prov:Association"
                            slots_with_inlined_classes[slot_name] = class_uri
                        else:
                            class_uri = f"{view.schema.default_prefix}:{element.name}"
                            slots_with_inlined_classes[slot_name] = class_uri

        # Populate the "collection-to-transformers" map for this specific migration.
        agenda = dict(
            biosample_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Biosample", slots_with_inlined_classes)],
            study_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Study", slots_with_inlined_classes)],
            extraction_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Extraction", slots_with_inlined_classes)],
            mags_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MagsAnalysis", slots_with_inlined_classes)],
            metabolomics_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetabolomicsAnalysis", slots_with_inlined_classes)],
            metaproteomics_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetaproteomicsAnalysis", slots_with_inlined_classes)],
            data_generation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:DataGeneration", slots_with_inlined_classes)],
            read_qc_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:ReadQcAnalysis", slots_with_inlined_classes)],
            library_preparation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:LibraryPreparation", slots_with_inlined_classes)],
            pooling_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Pooling", slots_with_inlined_classes)],
            metagenome_annotation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetagenomeAnnotation", slots_with_inlined_classes)],
            metagenome_assembly_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetagenomeAssembly", slots_with_inlined_classes)],
            metagenome_sequencing_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetagenomeSequencing", slots_with_inlined_classes)],
            nom_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:NomAnalysis", slots_with_inlined_classes)],
            read_based_taxonomy_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:ReadBasedTaxonomyAnalysis", slots_with_inlined_classes)],
            metatranscriptome_annotation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetatranscriptomeAnnotation", slots_with_inlined_classes)],
            metatranscriptome_assembly_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetatranscriptomeAssembly", slots_with_inlined_classes)],
            metatranscriptome_expression_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetatranscriptomeExpressionAnalysis", slots_with_inlined_classes)],
            collecting_biosamples_from_site_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:CollectingBiosamplesFromSite", slots_with_inlined_classes)]
        )  
        

        for collection_name, pipeline in agenda.items():
            self.adapter.process_each_document(collection_name=collection_name, pipeline=pipeline)

        # For each collection that is not allowed to contain documents having any slots having inlined classes,
        # we can use this more optimized adapter method to set the `type` field of all documents in that collection.
        self.adapter.set_field_of_each_document("data_object_set", "type", "nmdc:DataObject")
        self.adapter.set_field_of_each_document("functional_annotation_agg", "type", "nmdc:FunctionalAnnotationAggMember")
        self.adapter.set_field_of_each_document("field_research_site_set", "type", "nmdc:FieldResearchSite")
        self.adapter.set_field_of_each_document("processed_sample_set", "type", "nmdc:ProcessedSample")


    def add_type_to_inlined_classes(self, document: dict, slot: str, uri: str):
        r"""
        Adds a type slot to each inlined instance of an NMDC class in the biosample_set, study_set, extraction_set, 
        mags_set, metabolomics_analysis_set, and metaproteomics_analysis_set. 
        """
        if document.get(slot):
            # If slot is a list, iterate over each item in the list (e.g. chem_administration and has_credit_associations)
            if isinstance(document[slot], list):
                for item in document[slot]:
                    # will exclude non-inlined class slots that use an id
                    if not isinstance(item, str):
                        item["type"] = uri
                        if item.get("term"):
                            item["term"]["type"] = "nmdc:OntologyClass"
                        if item.get("applies_to_person"):
                            item["applies_to_person"]["type"] = "nmdc:PersonValue"
            else:
                # If slot is not multivalued
                document[slot]["type"] = uri
                if document[slot].get("term"):
                    document[slot]["term"]["type"] = "nmdc:OntologyClass"
                    

    def add_type_slot_with_class_uri(self, document: dict, class_uri: str, inlined_slots: Optional[dict] = None):
        r"""
            Adds a type slot to each collection with the appropriate class uri as the value. E.g. type: nmdc:Biosample. If a type
            slot exists, it will overwrite to the types listed below. If the optional dictionary argument inlined_slots is given, then
            this function will add the type lost to any slots that have inlined classes as their range.
        
            >>> m = Migrator()
            >>> m.add_type_slot_with_class_uri({'id': 123, 'collection_date': {'has_raw_value': '2017-05-09'}}, 'nmdc:Biosample', {'collection_date': 'nmdc:TimestampValue'}) 
            {'id': 123, 'collection_date': {'has_raw_value': '2017-05-09', 'type': 'nmdc:TimestampValue'}, 'type': 'nmdc:Biosample'}
            >>> m.add_type_slot_with_class_uri({'id': 567, 'type': 'nmdc:DataGeneration'}, 'nmdc:DataGeneration')
            {'id': 567, 'type': 'nmdc:DataGeneration'}
            >>> m.add_type_slot_with_class_uri({'id': 567, 'env_broad_scale': {'term': {'id': 'ENVO:1234'}}}, 'nmdc:Biosample', {'env_broad_scale': 'nmdc:ControlledIdentifiedTermValue'})
            {'id': 567, 'env_broad_scale': {'term': {'id': 'ENVO:1234', 'type': 'nmdc:OntologyClass'}, 'type': 'nmdc:ControlledIdentifiedTermValue'}, 'type': 'nmdc:Biosample'}
            >>> m.add_type_slot_with_class_uri({'id': 456}, 'nmdc:NomAnalysis')
            {'id': 456, 'type': 'nmdc:NomAnalysis'}
        """
        
        # Adds the type slot with the correct class_uri as a value to each collection instance
        document["type"] = class_uri
                
        # Add the type slot to any inlined classes in the Biosample, Study, or Extraction collections
        if inlined_slots:
            for slot, uri in inlined_slots.items():
                self.add_type_to_inlined_classes(document, slot, uri)
                
        return document

    


from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.10.0.part_1"
    _to_version = "11.10.0.part_2"

    def upgrade(self, commit_changes: bool = True) -> None:
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        
        Note: The `commit_changes` parameter is not used, but must be present in the
              method signature to comply with the base class. This migrator does
              not use transactions—all changes are committed immediately.

        TODO: Address the fact that existing migrators' `upgrade` methods
              no longer comply with the base class.
        """

        self.adapter.process_each_document(
            "workflow_execution_set", [self.update_processing_metadata]
        )

    def update_processing_metadata(self, workflow_execution: dict) -> dict:
        r"""
        Update the document's processing_institution and execution_resource values.

        >>> m = Migrator()
        >>> m.update_processing_metadata({'id': 1, 'execution_resource': 'JGI'})
        {'id': 1, 'processing_institution': 'JGI'}
        >>> m.update_processing_metadata({'id': 1, 'execution_resource': 'EMSL'})
        {'id': 1, 'processing_institution': 'NMDC'}
        >>> m.update_processing_metadata({'id': 1, 'execution_resource': 'NERSC-Perlmutter'})
        {'id': 1, 'execution_resource': 'NERSC-Perlmutter', 'processing_institution': 'NMDC'}
        >>> m.update_processing_metadata({'id': 1, 'execution_resource': 'EMSL-RZR'})
        {'id': 1, 'execution_resource': 'EMSL-RZR', 'processing_institution': 'NMDC'}
        >>> m.update_processing_metadata({'id': 1, 'execution_resource': 'NERSC-Cori'})
        {'id': 1, 'execution_resource': 'NERSC-Cori', 'processing_institution': 'NMDC'}

        """
        self.logger.info(f"Processing WorkflowExecution: {workflow_execution['id']}")
        if workflow_execution["execution_resource"] == "JGI":
            workflow_execution["processing_institution"] = "JGI"
            workflow_execution.pop("execution_resource", None)
        elif workflow_execution["execution_resource"] == "EMSL":
            workflow_execution["processing_institution"] = "NMDC"
            workflow_execution.pop("execution_resource", None)
        elif workflow_execution["execution_resource"] == "NERSC-Perlmutter":
            workflow_execution["processing_institution"] = "NMDC"
        elif workflow_execution["execution_resource"] == "EMSL-RZR":
            workflow_execution["processing_institution"] = "NMDC"
        elif workflow_execution["execution_resource"] == "NERSC-Cori":
            workflow_execution["processing_institution"] = "NMDC"
        else:
             raise ValueError(
                 f"No migration action defined for {workflow_execution['id']}"
             )
        return workflow_execution



import logging
from typing import Any, List, Dict
from importlib import resources
from pathlib import Path
from functools import lru_cache
import yaml
from nmdc_schema.nmdc_data import get_nmdc_schema_definition
from linkml_runtime import SchemaView

logger = logging.getLogger(__name__)

# TODO: Determine directory path and package path dynamically, instead of hard-coding them.
MIGRATOR_DIRECTORY = "nmdc_schema/migrators"
MIGRATOR_PACKAGE = "nmdc_schema.migrators"


def load_yaml_asset(path_to_asset_file: str) -> Any:
    """
    Safely loads the contents of a YAML file into a Python object. This function was designed specifically to
    work in both the `nmdc-schema` development environment and in environments that import the `nmdc-schema` package.

    Parameters:
        path_to_asset_file: Filesystem path to the YAML file, relative to the root of the `assets` directory.

    References:
    - https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.joinpath
    - https://docs.python.org/3/library/importlib.resources.html#importlib.resources.files
    - https://docs.python.org/3/library/importlib.resources.html#importlib.resources.as_file
    - https://realpython.com/python-yaml/#load-a-document-from-a-string-a-file-or-a-stream
    """
    # Define the Python path someone could use to `import` the package containing the `assets` directory.
    package_import_path = MIGRATOR_PACKAGE

    # Define the filesystem path to the `assets` directory (relative to that package).
    path_to_assets_directory = "assets"

    # Build a filesystem path to the specified asset file (also relative to that package).
    path_to_asset = Path(path_to_assets_directory).joinpath(path_to_asset_file)

    # Create a `Traversable` object that can be passed to the `resources.as_file()` function.
    traversable_resource_container = resources.files(package_import_path).joinpath(path_to_asset)

    # Use `resources.as_file()` to get an `open()`-able path to the asset file.
    with resources.as_file(traversable_resource_container) as file_path:
        # Finally, open the asset file and parse its contents as YAML.
        with open(file_path, "r") as f:
            obj = yaml.safe_load(f)  # the object type depends upon the YAML data

    return obj


def create_schema_view() -> SchemaView:
    """
    Returns a LinkML SchemaView instance that can be used to traverse the schema.
    
    >>> isinstance(create_schema_view(), SchemaView)
    True
    """
    schema_definition = get_nmdc_schema_definition()
    schema_view = SchemaView(schema_definition)
    return schema_view


# Constants for schema traversal
DATABASE_CLASS_NAME = "Database"


def get_classes_with_slots_by_range(schema_view: SchemaView, range_constraint: str) -> dict:
    """
    Returns a dictionary mapping class names to lists of their slot names that have the specified range.
    For each class, includes slots from the class AND all its subclasses to handle polymorphic storage.
    
    Args:
        schema_view: SchemaView instance
        range_constraint: The range type to search for (e.g., "QuantityValue", "TextValue", etc.)
        
    Returns:
        dict: {class_name: [slot_names_with_specified_range]}
    """
    classes_with_slots = {}
    
    # Get all classes in the schema
    for class_name in schema_view.all_classes():
        class_def = schema_view.get_class(class_name)
        if class_def:
            # Get all slots for this class
            induced_slots = schema_view.class_induced_slots(class_name)
            matching_slots = set()
            
            for slot_def in induced_slots:
                if slot_def.range == range_constraint:
                    matching_slots.add(slot_def.name)
            
            # Also get matching slots from all subclasses
            # This handles polymorphic storage where subclass records are stored in parent collections in mongodb
            subclass_slots = set()
            try:
                descendants = schema_view.class_descendants(class_name)
                for subclass_name in descendants:
                    subclass_induced_slots = schema_view.class_induced_slots(subclass_name)
                    for slot_def in subclass_induced_slots:
                        if slot_def.range == range_constraint:
                            subclass_slots.add(slot_def.name)
            except (AttributeError, KeyError, TypeError):
                # If class_descendants method doesn't exist or class is invalid, skip subclass processing
                pass
            
            # Combine parent and subclass slots
            all_slots = set(matching_slots) | subclass_slots
            
            if all_slots:
                classes_with_slots[class_name] = list(all_slots)
    
    return classes_with_slots


# TODO: Check whether this function can be replaced by a similar function from `refscan`.
def get_database_collections_for_class(schema_view: SchemaView, class_name: str) -> List[str]:
    """
    Returns the list of database collection names associated with the specified class.
    This function retrieves the collection names from the slots of the Database class
    that have the specified class as their range.

    Args:
        schema_view: SchemaView instance
        class_name: The name of the class to find collections for

    Returns:
        List[str]: List of collection names which permit storage of instances of the specified class
    """
    collection_names = set()

    for database_slot_name in schema_view.class_slots(DATABASE_CLASS_NAME):
        database_slot_def = schema_view.get_slot(database_slot_name, strict=True)
        database_slot_range_descendants = schema_view.class_descendants(database_slot_def.range, reflexive=True)
        if class_name in database_slot_range_descendants:
            collection_names.add(database_slot_name)

    return list(collection_names)


from nmdc_schema.migrators.adapters.adapter_base import AdapterBase
from nmdc_schema.migrators.adapters.mongo_adapter import MongoAdapter
from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.helpers import create_schema_view, logger
from nmdc_schema.migrators.migration_reporter import (
    MigrationReporter, 
    get_most_specific_class_for_reporting,
    parse_schema_path,
    get_clean_schema_path,
    resolve_class_from_schema_path
)
from typing import Optional, List, Any
from functools import lru_cache
import logging

# Constants for schema traversal
DATABASE_CLASS_NAME = "Database"
# Configure logging and initialize generic reporter
logging.basicConfig(level=logging.INFO, format='%(message)s')
logger.setLevel(logging.INFO)

@lru_cache
def get_collection_names_with_qv_slots_from_schema() -> List[str]:
    """
    Returns the names of the slots of the `Database` class that describe database collections
    and whose range classes could potentially contain QuantityValue objects.

    This method is vendored in from nmdc-runtime because we would rather not introduce a circular dependency.
    # TODO: consider moving migrators to runtime.

    Source: https://github.com/microbiomedata/refscan/blob/af092b0e068b671849fe0f323fac2ed54b81d574/refscan/lib/helpers.py#L31
    """
    collection_names = []

    schema_view = create_schema_view()
    for slot_name in schema_view.class_slots(DATABASE_CLASS_NAME):
        slot_definition = schema_view.induced_slot(slot_name, DATABASE_CLASS_NAME)

        if slot_definition.multivalued and slot_definition.inlined_as_list:
            # Only include collections whose range classes could contain QuantityValue objects
            if _collection_could_contain_quantity_values(schema_view, slot_definition.range):
                collection_names.append(slot_name)

    collection_names = list(set(collection_names))

    return collection_names


def _collection_could_contain_quantity_values(schema_view, range_class: str) -> bool:
    """
    Determines if a collection's range class could contain QuantityValue objects.
    
    Args:
        schema_view: SchemaView instance
        range_class: The range class name (e.g., "Biosample", "MaterialProcessing")
        
    Returns:
        bool: True if the class or any of its subclasses have QuantityValue slots
    """

    # Check if this class has any QuantityValue slots
    class_slots = schema_view.class_induced_slots(range_class)
    for slot_def in class_slots:
        if slot_def.range == "QuantityValue":
            return True

        # Check if any of its descendants have QuantityValue slots
        descendants = schema_view.class_descendants(range_class)
        for desc_class in descendants:
            desc_slots = schema_view.class_induced_slots(desc_class)
            for slot_def in desc_slots:
                if slot_def.range == "QuantityValue":
                    return True
    return False


class Migrator(MigratorBase):
    r"""
    Migrates a database between schemas 11.9.1 and 11.10.0.
    
    This migrator ensures that all QuantityValue instances have proper UCUM units.
    It handles missing units, normalizes existing units, and applies special case conversions.
    
    Examples:
    
    >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
    >>> import logging
    >>> logging.basicConfig(level=logging.WARNING)  # Suppress output for doctest
    >>> 
    >>> # Test missing unit addition
    >>> database = {
    ...     "biosample_set": [
    ...         {
    ...             "id": "nmdc:bsm-test1",
    ...             "type": "nmdc:Biosample", 
    ...             "temp": {"type": "nmdc:QuantityValue", "has_raw_value": "25"}
    ...         }
    ...     ]
    ... }
    >>> m = Migrator(DictionaryAdapter(database))
    >>> # Initialize required dependencies for standalone testing
    >>> from nmdc_schema.migrators.helpers import create_schema_view
    >>> from nmdc_schema.migrators.migration_reporter import MigrationReporter
    >>> from nmdc_schema.migrators.migration_reporter import get_most_specific_class_for_reporting
    >>> from nmdc_schema.migrators.migration_reporter import parse_schema_path
    >>> from nmdc_schema.migrators.migration_reporter import get_clean_schema_path
    >>> from nmdc_schema.migrators.migration_reporter import resolve_class_from_schema_path
    >>> m._schema_view = create_schema_view()
    >>> m._unit_alias_map = m._build_unit_alias_map(m._schema_view)
    >>> m.reporter = MigrationReporter(m.logger)
    >>> doc = m.ensure_quantity_value_has_unit(database["biosample_set"][0])
    >>> doc["temp"]["has_unit"]
    'Cel'
    
    >>> # Test unit normalization (Celsius -> Cel)
    >>> database2 = {
    ...     "biosample_set": [
    ...         {
    ...             "id": "nmdc:bsm-test2",
    ...             "type": "nmdc:Biosample",
    ...             "temp": {"type": "nmdc:QuantityValue", "has_raw_value": "30", "has_unit": "Celsius"}
    ...         }
    ...     ]
    ... }
    >>> m2 = Migrator(DictionaryAdapter(database2))
    >>> # Initialize required dependencies for standalone testing
    >>> m2._schema_view = create_schema_view()
    >>> m2._unit_alias_map = m2._build_unit_alias_map(m2._schema_view)
    >>> m2.reporter = MigrationReporter(m2.logger)
    >>> doc2 = m2.ensure_quantity_value_has_unit(database2["biosample_set"][0])
    >>> doc2["temp"]["has_unit"]
    'Cel'
    
    >>> # Test carb_nitro_ratio special case (dimensionless)
    >>> database3 = {
    ...     "biosample_set": [
    ...         {
    ...             "id": "nmdc:bsm-test3", 
    ...             "type": "nmdc:Biosample",
    ...             "carb_nitro_ratio": {"type": "nmdc:QuantityValue", "has_raw_value": "12.5"}
    ...         }
    ...     ]
    ... }
    >>> m3 = Migrator(DictionaryAdapter(database3))
    >>> # Initialize required dependencies for standalone testing
    >>> m3._schema_view = create_schema_view()
    >>> m3._unit_alias_map = m3._build_unit_alias_map(m3._schema_view)
    >>> m3.reporter = MigrationReporter(m3.logger)
    >>> doc3 = m3.ensure_quantity_value_has_unit(database3["biosample_set"][0])
    >>> doc3["carb_nitro_ratio"]["has_unit"]
    '1'
    
    >>> # Test nested QuantityValue (PortionOfSubstance)
    >>> database4 = {
    ...     "biosample_set": [
    ...         {
    ...             "id": "nmdc:bsm-test4",
    ...             "type": "nmdc:Biosample",
    ...             "substances_used": [
    ...                 {
    ...                     "type": "nmdc:PortionOfSubstance",
    ...                     "volume": {"type": "nmdc:QuantityValue", "has_raw_value": "10"}
    ...                 }
    ...             ]
    ...         }
    ...     ]
    ... }
    >>> m4 = Migrator(DictionaryAdapter(database4))
    >>> # Initialize required dependencies for standalone testing
    >>> m4._schema_view = create_schema_view()
    >>> m4._unit_alias_map = m4._build_unit_alias_map(m4._schema_view)
    >>> m4.reporter = MigrationReporter(m4.logger)
    >>> doc4 = m4.ensure_quantity_value_has_unit(database4["biosample_set"][0])
    >>> doc4["substances_used"][0]["volume"]["has_unit"]
    'mL'
    """

    _from_version = "11.9.1"
    _to_version = "11.10.0.part_1"

    # Mapping of class/slot combinations to their appropriate UCUM units
    QUANTITY_VALUE_UNITS = {
        "nmdc:PortionOfSubstance": {
            "final_concentration": "%",
            "volume": "mL"
        },
        "nmdc:Biosample": {
            "humidity": "%",
            "tot_carb": "%",
            "tot_nitro_content": "%",
            "samp_store_temp": "Cel",
            "temp": "Cel",
            "avg_temp": "Cel",
            "host_height": "cm",
            "host_age": "d",
            "host_dry_mass": "g",
            "samp_size": "g",
            "abs_air_humidity": "kPa",
            "depth": "m",
            "wind_speed": "m/s",
            "subsurface_depth": "m",
            "ammonium_nitrogen": "mg/kg",
            "calcium": "mg/kg",
            "magnesium": "mg/kg",
            "manganese": "mg/kg",
            "nitrate_nitrogen": "mg/kg",
            "nitrite_nitrogen": "mg/g",
            "potassium": "mg/kg",
            "zinc": "mg/kg",
            "ammonium": "mg/L",
            "chloride": "mg/L",
            "diss_inorg_carb": "mg/L",
            "diss_inorg_nitro": "mg/L",
            "diss_iron": "mg/L",
            "diss_org_carb": "mg/L",
            "diss_oxygen": "mg/L",
            "sodium": "mg/L",
            "soluble_react_phosp": "mg/L",
            "sulfate": "mg/L",
            "tot_org_carb": "mg/L",
            "tot_phosp": "mg/L",
            "nitro": "%",
            "org_carb": "%",
            "tot_nitro": "%",
            "lbc_thirty": "[ppm]",
            "lbceq": "[ppm]",
            "photon_flux": "[arb'U]{micro_Einsteins}/m2/s",
            "conduc": "uS/cm",
            "solar_irradiance": "W/m2",
            "chlorophyll": "micromol/L"
        },
        "nmdc:ChemicalConversionProcess": {
            "temperature": "Cel",
            "duration": "h"
        },
        "nmdc:ChromatographyConfiguration": {
            "temperature": "Cel"
        },
        "nmdc:Extraction": {
            "input_mass": "g"
        },
        "nmdc:SubSamplingProcess": {
            "mass": "g"
        },
        "nmdc:MobilePhaseSegment": {
            "duration": "min"
        }
    }

    def __init__(self, adapter: Optional[AdapterBase] = None, logger=None):
        super().__init__(adapter, logger)
        self._unit_alias_map = None
        self.reporter: Optional[MigrationReporter] = None
        self._schema_view = None  # Cache schema view to avoid repeated creation

    def upgrade(self, commit_changes: bool = False) -> None:
        """
        Migrates all QuantityValue instances in records to have non-null has_unit values conformant to enumeration PVs.

        All operations are wrapped in a MongoDB transaction for rollback capability.
        All actions are logged in a reporter class so that we can see some statistics at the end of the migration.
        
        Args:
            commit_changes: If True, commits the transaction. If False (default), rolls back the transaction.
        """
        self.reporter = MigrationReporter(self.logger)
        
        # Get schema view to find all classes and their slots with QuantityValue range
        self._schema_view = create_schema_view()
        
        # Build unit alias map from UnitEnum (an enumeration of possible UCUM units that we use to harmonize units
        # that already exist in the database).
        self._unit_alias_map = self._build_unit_alias_map(self._schema_view)
        
        # Get the actual collection names from the Database class slots
        real_collection_names = get_collection_names_with_qv_slots_from_schema()
        
        # Warn user if commit_changes parameter is being ignored
        self._warn_if_commit_ignored(commit_changes)
        
        # Use adapter's transaction-aware processing method
        if isinstance(self.adapter, MongoAdapter):
            # MongoDB adapter - use transaction support
            try:
                self.adapter.process_collections_in_transaction(
                    collection_names=real_collection_names,
                    document_processor=self.ensure_quantity_value_has_unit,
                    commit_changes=False  # Always process without committing first
                )
                
                self.reporter.generate_final_report()
                
                # Validate migration success - raises exception if issues found
                self.reporter.validate_migration_success()
                
                # If validation passes and commit requested, commit the transaction
                if commit_changes:
                    # Re-run with commit=True since validation passed
                    self.adapter.process_collections_in_transaction(
                        collection_names=real_collection_names,
                        document_processor=self.ensure_quantity_value_has_unit,
                        commit_changes=True
                    )
                    self.logger.info("Migration validation passed. Transaction committed (changes have been saved)")
                else:
                    self.logger.info("Migration validation passed. Transaction rolled back (no changes were committed)")
                    
            except Exception as e:
                # Log rollback message with commit context
                if commit_changes:
                    self.logger.info("Migration validation failed. Transaction automatically rolled back (no changes were saved, despite commit request)")
                else:
                    self.logger.info("Migration validation failed. Transaction rolled back (no changes were committed)")
                self.logger.error(f"Migration failed: {e}")
                raise
        else:
            # Non-MongoDB adapter - process collections directly without transactions
            try:
                for collection_name in real_collection_names:
                    self.adapter.process_each_document(collection_name, [self.ensure_quantity_value_has_unit])
                
                self.reporter.generate_final_report()
                
                # Validate migration success - raises exception if issues found
                self.reporter.validate_migration_success()
                
                if not commit_changes:
                    self.logger.info("Migration validation passed. Note: Non-MongoDB adapter doesn't support rollback - changes are applied immediately")
                else:
                    self.logger.info("Migration validation passed.")
                    
            except Exception as e:
                # from S&E - raise exceptions if the has_unit can not be set.
                # both for when it can't map, and when it can't assign or has_unit
                self.logger.error(f"Migration failed: {e}")
                raise
    

    @staticmethod
    def _get_classes_with_quantity_value_slots(view) -> dict:
        r"""
        Returns a dictionary mapping class names to lists of their slot names that have QuantityValue range.
        For each class, includes slots from the class AND all its subclasses to handle polymorphic storage.
        
        Args:
            view: SchemaView instance
            
        Returns:
            dict: {class_name: [slot_names_with_quantity_value_range]}
        """
        classes_with_quantity_value_slots = {}
        
        # Get all classes in the schema
        for class_name in view.all_classes():
            class_def = view.get_class(class_name)
            if class_def:
                # Get all slots for this class
                induced_slots = view.class_induced_slots(class_name)
                quantity_value_slots = set()
                
                for slot_def in induced_slots:
                    if slot_def.range == "QuantityValue":
                        quantity_value_slots.add(slot_def.name)
                
                # Also get QuantityValue slots from all subclasses
                # This handles polymorphic storage where subclass records are stored in parent collections in mongodb
                subclass_slots = set()
                try:
                    descendants = view.class_descendants(class_name)
                    for subclass_name in descendants:
                        subclass_induced_slots = view.class_induced_slots(subclass_name)
                        for slot_def in subclass_induced_slots:
                            if slot_def.range == "QuantityValue":
                                subclass_slots.add(slot_def.name)
                except:
                    # If class_descendants doesn't exist, skip subclass processing
                    pass
                
                # Combine parent and subclass slots
                all_slots = set(quantity_value_slots) | subclass_slots
                
                if all_slots:
                    classes_with_quantity_value_slots[class_name] = list(all_slots)
        
        return classes_with_quantity_value_slots

    @staticmethod
    def _build_unit_alias_map(view) -> dict:
        r"""
        Builds a mapping from unit aliases to their canonical UCUM values using the schema.
        
        Args:
            view: SchemaView instance
            
        Returns:
            dict: {alias -> canonical_unit} mapping
        """
        alias_to_canonical = {}
        
        # Get UnitEnum from the provided schema view
        unit_enum = view.get_enum('UnitEnum')
        
        if unit_enum and unit_enum.permissible_values:
            for canonical_unit, unit_def in unit_enum.permissible_values.items():
                # The canonical unit maps to itself
                alias_to_canonical[canonical_unit] = canonical_unit
                
                # Add aliases that map to canonical unit
                if hasattr(unit_def, 'aliases') and unit_def.aliases:
                    for alias in unit_def.aliases:
                        alias_to_canonical[alias] = canonical_unit
        
        return alias_to_canonical

    def ensure_quantity_value_has_unit(self, document: dict) -> dict:
        r"""
        Ensures that all QuantityValue instances in a document have non-null has_unit values.
        
        This function recursively traverses the document to find all QuantityValue instances
        and adds appropriate units based on the class/slot mapping.
        
        Args:
            document (dict): A document from the database
            
        Returns:
            dict: The modified document with has_unit values added to QuantityValue instances
        """
        
        # Recursively traverse the entire document to find all QuantityValue instances
        self._traverse_and_fix_quantity_values(document, document_root=document)
        
        return document
    
    def _traverse_and_fix_quantity_values(self,
                                          obj: Any,
                                          document_root: dict,
                                          path: str = "") -> None:
        """
        Recursively traverses an object to find and fix QuantityValue instances.
        
        Args:
            obj: The object to traverse (dict, list, or other)
            document_root: The root document for context (e.g., to get document type)
            path: Current path in the document (for debugging)
        """
        if isinstance(obj, dict):
            # Check if this is a QuantityValue instance
            if obj.get('type') == 'nmdc:QuantityValue':
                # This is a QuantityValue, try to add/fix its unit
                self._fix_quantity_value_unit(obj, document_root, path)
            else:
                # Quick optimization: only recurse into values that could contain QuantityValue objects
                # Skip simple string/number values and common metadata fields
                for key, value in obj.items():
                    if key in ['_id', 'id', 'name', 'description', 'type'] and isinstance(value, str):
                        continue  # Skip simple string metadata
                    if isinstance(value, (int, float, bool)) or value is None:
                        continue  # Skip simple scalar values
                    
                    new_path = f"{path}.{key}" if path else key
                    # Recursively traverse into nested objects to find QuantityValue instances
                    # value: the nested object/dict at this key (e.g., {"type": "nmdc:QuantityValue", "has_raw_value": "25"})
                    # document_root: original document for context (e.g., {"id": "biosample1", "type": "nmdc:Biosample", "temp": {...}})
                    # new_path: current location path (e.g., "temp" or "substances_used[0].volume")
                    self._traverse_and_fix_quantity_values(value, document_root, new_path)
        elif isinstance(obj, list):
            # Recurse into list items
            for i, item in enumerate(obj):
                new_path = f"{path}[{i}]"
                self._traverse_and_fix_quantity_values(item, document_root, new_path)
    
    def _fix_quantity_value_unit(self, quantity_value: dict, document_root: dict, path: str) -> None:
        r"""
        Attempts to fix a QuantityValue instance by adding or normalizing its unit.
        
        Args:
            quantity_value: The QuantityValue instance to fix
            document_root: The full document for context
            path: Path to this QuantityValue in the document
        """
        # Get root collection class for reporting (the class that has a MongoDB collection)
        root_collection_class = document_root.get('type', 'nmdc:Unknown')
        # Get clean schema path without array indices for reporting
        clean_schema_path = get_clean_schema_path(path)
        
        # Check if `has_unit` is missing or is None
        if 'has_unit' not in quantity_value or quantity_value['has_unit'] is None:
            # Get most specific class for unit lookup
            most_specific_class = get_most_specific_class_for_reporting(self._schema_view, document_root, path)
            
            # Check for special cases where we can extract unit from raw_value
            unit = self._handle_one_off_unit_cases(quantity_value, most_specific_class, path, None)
            
            # If no special case, try to infer unit from document type and path
            if not unit:
                unit = self._infer_unit_from_context(document_root, path)
            
            if unit:
                quantity_value['has_unit'] = unit
                # Track record update: missing → unit
                self.reporter.track_record_updated(
                    class_name=root_collection_class,
                    slot_name=clean_schema_path, 
                    subclass_type=most_specific_class,
                    source_value="<missing>",
                    target_value=unit
                )
            else:
                # Report missing unit
                self.reporter.track_missing_value(root_collection_class, clean_schema_path)
        else:
            # has_unit exists, check if it needs normalization
            current_unit = quantity_value['has_unit']
            # Get most specific class for unit lookup (for special cases)
            most_specific_class = get_most_specific_class_for_reporting(self._schema_view, document_root, path)
            
            # Check if current unit is an alias that should be normalized
            if current_unit in self._unit_alias_map:
                canonical_unit = self._unit_alias_map[current_unit]
                
                # Only update if the canonical form is different
                if current_unit != canonical_unit:
                    quantity_value['has_unit'] = canonical_unit
                    # Track unit normalization
                    self.reporter.track_record_updated(
                        class_name=root_collection_class,
                        slot_name=clean_schema_path,
                        subclass_type=most_specific_class,
                        source_value=current_unit,
                        target_value=canonical_unit
                    )
                else:
                    # Unit is already in canonical form - track as conformant
                    self.reporter.track_record_processed(
                        class_name=root_collection_class,
                        slot_name=clean_schema_path,
                        subclass_type=most_specific_class,
                        value=current_unit
                    )
            else:
                # Check for special one-off cases
                handled_unit = self._handle_one_off_unit_cases(quantity_value, most_specific_class, path, current_unit)
                if handled_unit:
                    # One-off case was handled - track the change
                    self.reporter.track_record_updated(
                        class_name=root_collection_class,
                        slot_name=clean_schema_path,
                        subclass_type=most_specific_class,
                        source_value=current_unit or "<missing>",
                        target_value=handled_unit
                    )
                else:
                    # Unit is not in the alias map - report it for analysis
                    self.reporter.track_unmapped_value(root_collection_class, clean_schema_path, current_unit)
    
    def _infer_unit_from_context(self, full_document: dict, path: str) -> Optional[str]:
        r"""
        Infers the appropriate unit for a QuantityValue .
        
        Args:
            full_document: The full document for context
            path: Path to the QuantityValue in the document
            
        Returns:
            str or None: The inferred unit, or None if not found
        """
        # Parse path into components, filtering out array indices
        path_parts = parse_schema_path(path)
        if not path_parts:
            return None
        
        field_name = path_parts[-1]
        slot_path = path_parts[:-1]  # All parts except the final field name
        
        # Start with document's root class
        doc_type = full_document.get('type', 'nmdc:Unknown')
        root_class = doc_type.replace('nmdc:', '') if doc_type.startswith('nmdc:') else doc_type
        
        # Use schema to resolve the class context for this field
        target_class = resolve_class_from_schema_path(self._schema_view, root_class, slot_path)
        if target_class:
            return self._get_unit_for_class_slot(f"nmdc:{target_class}", field_name, None)
        
        # Fallback to document type if schema resolution fails
        return self._get_unit_for_class_slot(doc_type, field_name, None)
    
    def _add_unit_to_quantity_value(self, quantity_value: dict, class_uri: str, slot_name: str, full_document: Optional[dict] = None) -> None:
        r"""
        Adds an appropriate unit to a QuantityValue instance if it doesn't have one,
        or normalizes existing unit values using UnitEnum aliases.
        
        Args:
            quantity_value (dict): A QuantityValue instance that may be missing a has_unit field
            class_uri (str): The class URI (e.g., "nmdc:Biosample")
            slot_name (str): The slot name (e.g., "temp")
        """
        # Check if the slot has_unit is missing or is None
        if 'has_unit' not in quantity_value or quantity_value['has_unit'] is None:
            # Get the appropriate unit from the mapping
            # Check if document has a specific type that should be used for unit lookup
            lookup_class_uri = class_uri
            if full_document and 'type' in full_document:
                lookup_class_uri = full_document['type']
            
            unit = self._get_unit_for_class_slot(lookup_class_uri, slot_name, None)
            if unit:  # Only update if we found a valid mapping
                quantity_value['has_unit'] = unit
                # Track the unit addition using the generic reporter
                self.reporter.track_operation('units_added', unit, 1)
            
        else:
            # has_unit exists, check if it needs normalization
            current_unit = quantity_value['has_unit']
            
            # Check if current unit is an alias that should be normalized
            if current_unit in self._unit_alias_map:
                canonical_unit = self._unit_alias_map[current_unit]
                
                # Only update if the canonical form is different
                if current_unit != canonical_unit:
                    quantity_value['has_unit'] = canonical_unit
                    
                    # Track the unit normalization using the generic reporter
                    self.reporter.track_operation('units_normalized', f"{current_unit} → {canonical_unit}", 1)
            else:
                # Current unit is not in our alias map
                # Check for special one-off cases first
                if self._handle_one_off_unit_cases(quantity_value, class_uri, slot_name, current_unit):
                    # One-off case was handled, continue to next
                    pass
                elif current_unit in self._unit_alias_map and self._unit_alias_map[current_unit] == current_unit:
                    # Unit is already a valid canonical enum value, no action needed
                    pass
                else:
                    # Current unit is not a valid enum value, check if it can be mapped
                    # Check if document has a specific type that should be used for unit lookup
                    lookup_class_uri = class_uri
                    if full_document and 'type' in full_document:
                        lookup_class_uri = full_document['type']
                    
                    self._get_unit_for_class_slot(lookup_class_uri, slot_name, current_unit)
                    # If no mapping found, just report it but don't change the value
    
    def _handle_one_off_unit_cases(self, quantity_value: dict, class_uri: str, path: str, current_unit: Optional[str]) -> Optional[str]:
        """
        Handle special one-off unit conversion cases.

        In particular, this method handles cases where the unit is missing, and we have already looked at the records
        ahead of time to determine what the unit should be, or where we need to convert a specific unit.
        
        Args:
            quantity_value (dict): The QuantityValue instance to potentially modify
            class_uri (str): The class URI (e.g., "nmdc:Biosample")
            path (str): The full path to the QuantityValue field (e.g., "substances_used[0].volume")
            current_unit (str): The current unit value (None if missing)
            
        Returns:
            str or None: The extracted/converted unit, or None if not handled
        """
        # Extract the field name from the path
        slot_name = path.split('.')[-1] if path else ""
        # Handle missing units by extracting from has_raw_value
        if current_unit is None:
            # Special case: Biosample salinity - extract specific units from has_raw_value
            if class_uri == "nmdc:Biosample" and slot_name == "salinity":
                raw_value = quantity_value.get('has_raw_value')
                if raw_value and isinstance(raw_value, str):
                    extracted_unit = self._extract_salinity_unit(raw_value)
                    if extracted_unit:
                        return extracted_unit

            # Special case: Biosample carb_nitro_ratio - set unit to "1" (dimensionless)
            if class_uri == "nmdc:Biosample" and slot_name == "carb_nitro_ratio":
                return "1"
        
        # Handle existing units that need conversion
        else:
            # Special case: Biosample nitrate with "detection" unit should become "umol/L"
            if class_uri == "nmdc:Biosample" and slot_name == "nitrate" and current_unit == "detection":
                quantity_value['has_unit'] = "umol/L"
                return "umol/L"  # Return the unit to indicate it was handled

        return None
    
    def _extract_salinity_unit(self, raw_value: str) -> Optional[str]:
        """
        Extract specific salinity units from a raw value string.
        Only looks for: mgL → mg/L, mg/L → mg/L, % → %
        
        Args:
            raw_value: The raw value string that may contain salinity units
            
        Returns:
            str or None: The normalized unit, or None if not found
        """
        # Look for specific salinity units only
        if 'mgL' in raw_value:
            return 'mg/L'
        elif 'mg/L' in raw_value:
            return 'mg/L'  
        elif '%' in raw_value:
            return '%'
        
        return None
    
    def _get_unit_for_class_slot(self, class_uri: str, slot_name: str, current_unit_value: str = None) -> Optional[str]:
        r"""
        Gets the appropriate unit for a given class and slot combination.
        Also check parent classes in the inheritance hierarchy if no direct mapping is found.
        
        Args:
            class_uri (str): The class URI (e.g., "nmdc:Biosample")
            slot_name (str): The slot name (e.g., "temp")
            current_unit_value (str, optional): The current unit value that couldn't be mapped
            
        Returns:
            str or None: The appropriate unit, or None if no mapping is found

        """
        # Look up the unit in the mapping
        class_units = self.QUANTITY_VALUE_UNITS.get(class_uri, {})
        unit = class_units.get(slot_name)
        
        if unit:
            return unit
        
        # If not found, check parent classes in the inheritance hierarchy
        unit = self._get_unit_from_inheritance_chain(class_uri, slot_name)
        if unit:
            return unit
        
        # Only track as unmapped if the current unit value is not already valid
        if current_unit_value is not None and current_unit_value not in self._unit_alias_map:
            # Track unmapped slots and values using the generic reporter
            slot_key = f"{class_uri}.{slot_name}"
            self.reporter.track_item('unmapped_slots', slot_key)
            self.reporter.track_value_set('unmapped_values', slot_key, current_unit_value)
        
        # Return None instead of a fallback unit - just report and don't update
        return None
    
    def _get_unit_from_inheritance_chain(self, class_uri: str, slot_name: str) -> Optional[str]:
        r"""
        Checks parent classes in the inheritance hierarchy for unit mappings.
        
        Args:
            class_uri (str): The class URI (e.g., "nmdc:DissolvingProcess")
            slot_name (str): The slot name (e.g., "volume")
            
        Returns:
            str or None: The appropriate unit from a parent class, or None if not found
        """
        # Use cached schema view to traverse inheritance
        view = self._schema_view
        
        # Remove the "nmdc:" prefix to get the class name
        class_name = class_uri.replace('nmdc:', '')
        
        try:
            # Get the class definition
            class_def = view.get_class(class_name)
            if not class_def:
                return None
                
            # Get all ancestor classes
            ancestors = view.class_ancestors(class_name)
            
            # Check each ancestor for unit mappings
            for ancestor_name in ancestors:
                ancestor_uri = f"nmdc:{ancestor_name}"
                ancestor_units = self.QUANTITY_VALUE_UNITS.get(ancestor_uri, {})
                unit = ancestor_units.get(slot_name)
                if unit:
                    return unit
                    
        except Exception:
            # If anything goes wrong with schema traversal, just return None
            pass
            
        return None



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.6.0"
    _to_version = "11.6.1"

    def upgrade(self) -> None:
        r"""No upgrade needed."""
        pass



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.10.0.part_2"
    _to_version = "11.10.0.part_3"

    def upgrade(self, commit_changes: bool = True) -> None:
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        
        Note: The `commit_changes` parameter is not used, but must be present in the
              method signature to comply with the base class. This migrator does
              not use transactions—all changes are committed immediately.
        """

        self.adapter.process_each_document(
            "data_generation_set", [self.confirm_specific_fields_are_absent]
        )

    def confirm_specific_fields_are_absent(self, data_generation: dict) -> dict:
        r"""
        Raises an exception if the document represents a `NucleotideSequencing` and either
        its `target_gene` field or its `target_subfragment` field is present.

        Note: By raising an exception, we bring this "unexpected" situation to the attention
              of the person performing the migration, so they can take action (e.g., aborting
              the migration and contacting the schema change author).

        Reference: https://microbiomedata.github.io/nmdc-schema/NucleotideSequencing/

        >>> m = Migrator()

        # Test: Document has neither field or is of irrelevant type.
        >>> m.confirm_specific_fields_are_absent({"id": 1, "type": "nmdc:NucleotideSequencing"})
        {'id': 1, 'type': 'nmdc:NucleotideSequencing'}
        >>> m.confirm_specific_fields_are_absent({"id": 2, "type": "nmdc:Other", "target_gene": None})
        {'id': 2, 'type': 'nmdc:Other', 'target_gene': None}

        # Test: Document has either field.
        >>> m.confirm_specific_fields_are_absent({"id": 3, "type": "nmdc:NucleotideSequencing", "target_gene": None})
        Traceback (most recent call last):
            ...
        ValueError: NucleotideSequencing 3 has field(s): ['target_gene']
        >>> m.confirm_specific_fields_are_absent({"id": 4, "type": "nmdc:NucleotideSequencing", "target_subfragment": None})
        Traceback (most recent call last):
            ...
        ValueError: NucleotideSequencing 4 has field(s): ['target_subfragment']
        
        # Test: Document has both fields.
        >>> m.confirm_specific_fields_are_absent({"id": 5, "type": "nmdc:NucleotideSequencing", "target_gene": None, "target_subfragment": None})
        Traceback (most recent call last):
            ...
        ValueError: NucleotideSequencing 5 has field(s): ['target_gene', 'target_subfragment']
        """

        # Note: We only want to process `NucleotideSequencing` documents.
        if data_generation["type"] == "nmdc:NucleotideSequencing":

            # Give the variable a more specific name.
            nucleotide_sequencing = data_generation
            nucleotide_sequencing_id = nucleotide_sequencing["id"]

            # Check which of the fields are present.
            field_presence = {
                "target_gene": "target_gene" in nucleotide_sequencing,
                "target_subfragment": "target_subfragment" in nucleotide_sequencing,
            }

            # If either field is present, raise an exception.
            if any(field_presence.values()):
                present_field_names = [name for name, is_present in field_presence.items() if is_present]
                raise ValueError(
                    f"NucleotideSequencing {nucleotide_sequencing_id} has field(s): {present_field_names}"
                )

        return data_generation



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    """
    Migrates data from X to PR129, namely changes the slot names has_metabolite_quantifications to has_metabolite_identifications
    and metabolite_quantified to metabolite_identified.
    """

    _from_version = "X"
    _to_version = "PR129"

    def upgrade(self):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        """

        self.adapter.process_each_document(collection_name="metabolomics_analysis_activity_set", pipeline=[self.change_metab_slot_names])

    def change_metab_slot_names(self, metab_doc: dict) -> dict:
        r"""
        Changes the slot names has_metabolite_quantifications to has_metabolite_identifications
        and metabolite_quantified to metabolite_identified.

        >>> m = Migrator()  
        >>> m.change_metab_slot_names({'id': 123, 'has_metabolite_quantifications': [{'metabolite_quantified': 'chebi:16997'}]})  
        {'id': 123, 'has_metabolite_identifications': [{'metabolite_identified': 'chebi:16997'}]}
        """

        if "has_metabolite_quantifications" in metab_doc:
            metab_doc["has_metabolite_identifications"] = metab_doc.pop("has_metabolite_quantifications")
            for metabolite in metab_doc["has_metabolite_identifications"]:
                metabolite["metabolite_identified"] = metabolite.pop("metabolite_quantified")
           
        return metab_doc



import sys
from pathlib import Path

from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.adapters.mongo_adapter import MongoAdapter

project_root = Path(__file__).parent.parent.parent
sys.path.insert(0, str(project_root))

class Migrator(MigratorBase):
    r'''
    Migrates a database between two schemas.
    '''

    _from_version = '11.11.0'
    _to_version = '11.13.0.part_1'

    def upgrade(self, commit_changes: bool = False) -> None:
        r'''
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        '''

        if isinstance(self.adapter, MongoAdapter):
            try:
                self.adapter.execute_in_transaction(
                    operations_callback=self.perform_all_migration_operations,
                    commit_changes=commit_changes
                )

                if commit_changes:
                    self.logger.info("Transaction committed (changes have been saved)")
                else:
                    self.logger.info("Transaction rolled back (no changes were committed)")

            except Exception as e:
                self.logger.error(f"Migration failed: {e}")
                raise
        else:
            raise NotImplementedError("This migrator currently only supports MongoAdapter.")


    def perform_all_migration_operations(self, _, __) -> None:
        self.logger.info("Moving problematic QuantityValue entries to misc_param")
        self.adapter.process_each_document("biosample_set", [
            self.migrate_misc_param_to_PropertyAssertion_range,
            self.move_problematic_values_to_misc_param,
        ])


    def migrate_misc_param_to_PropertyAssertion_range(self, biosample: dict) -> dict:
        r"""The misc_param slot on Biosample used to have range TextValue, but now has range
        PropertyAssertion. This migrator changes misc_param values to conform to the new range.

        >>> m = Migrator()

        # Test: Biosample with no misc_param
        >>> biosample_no_misc = {
        ...     "id": "test1",
        ...     "type": "nmdc:Biosample"
        ... }
        >>> m.migrate_misc_param_to_PropertyAssertion_range(biosample_no_misc)
        {'id': 'test1', 'type': 'nmdc:Biosample'}

        # Test: Biosample with misc_param as TextValue
        >>> biosample_with_misc = {
        ...     "id": "test2",
        ...     "type": "nmdc:Biosample",
        ...     "misc_param": [{"type": "nmdc:TextValue", "has_raw_value": "biomass_yield;2.565699 Mg/ha"}]
        ... }
        >>> m.migrate_misc_param_to_PropertyAssertion_range(biosample_with_misc)
        {'id': 'test2', 'type': 'nmdc:Biosample', 'misc_param': [{'type': 'nmdc:PropertyAssertion', 'has_raw_value': 'biomass_yield;2.565699 Mg/ha', 'has_attribute_label': 'biomass_yield', 'has_numeric_value': 2.565699, 'has_unit': 'Mg/ha'}]}
        """
        misc_param = biosample.get("misc_param")
        if not misc_param:
            return biosample

        property_assertions = []
        for item in misc_param:
            if isinstance(item, dict) and item.get("type") == "nmdc:TextValue":
                # Convert TextValue to PropertyAssertion
                raw_value = item.get("has_raw_value")
                if not raw_value:
                    continue
                pa = {
                    "type": "nmdc:PropertyAssertion",
                    "has_raw_value": raw_value
                }
                if ";" in raw_value:
                    [label, _, value] = item.get("has_raw_value", "").partition(";")
                    pa["has_attribute_label"] = label.strip()
                    if " " in value:
                        num_str, _, unit = value.strip().partition(" ")
                        try:
                            pa["has_numeric_value"] = float(num_str)
                        except ValueError:
                            pass  # If conversion fails, skip adding has_numeric_value
                        if unit:
                            pa["has_unit"] = unit.strip()
                property_assertions.append(pa)
            else:
                raise ValueError(
                    f"Biosample {biosample.get('id')} has misc_param item that is not a TextValue: {item}"
                )
        biosample["misc_param"] = property_assertions
        return biosample


    def move_problematic_values_to_misc_param(self, biosample: dict) -> dict:
        r"""Move QuantityValue entries with invalid units to misc_param as PropertyAssertion.

        >>> m = Migrator()

        # Test: Biosample with valid QuantityValue entries
        >>> biosample_valid = {
        ...     "id": "test1",
        ...     "type": "nmdc:Biosample",
        ...     "abs_air_humidity": {
        ...         "type": "nmdc:QuantityValue",
        ...         "has_unit": "kg/kg",
        ...         "has_numeric_value": 75.0
        ...     },
        ...     "diss_oxygen": {
        ...         "type": "nmdc:QuantityValue",
        ...         "has_unit": "umol/L",
        ...         "has_numeric_value": 8.0
        ...     },
        ...     "solar_irradiance": {
        ...         "type": "nmdc:QuantityValue",
        ...         "has_unit": "kW/m2/d",
        ...         "has_numeric_value": 200.0
        ...     }
        ... }
        >>> m.move_problematic_values_to_misc_param(biosample_valid)
        {'id': 'test1', 'type': 'nmdc:Biosample', 'abs_air_humidity': {'type': 'nmdc:QuantityValue', 'has_unit': 'kg/kg', 'has_numeric_value': 75.0}, 'diss_oxygen': {'type': 'nmdc:QuantityValue', 'has_unit': 'umol/L', 'has_numeric_value': 8.0}, 'solar_irradiance': {'type': 'nmdc:QuantityValue', 'has_unit': 'kW/m2/d', 'has_numeric_value': 200.0}}

        # Test: Biosample with problematic QuantityValue entries
        >>> biosample_problematic = {
        ...     "id": "test2",
        ...     "type": "nmdc:Biosample",
        ...     "abs_air_humidity": {
        ...         "type": "nmdc:QuantityValue",
        ...         "has_unit": "kPa",
        ...         "has_numeric_value": 75.0
        ...     },
        ...     "diss_oxygen": {
        ...         "type": "nmdc:QuantityValue",
        ...         "has_unit": "mL/L",
        ...         "has_numeric_value": 8.0
        ...     },
        ...     "solar_irradiance": {
        ...         "type": "nmdc:QuantityValue",
        ...         "has_unit": "W/m2",
        ...         "has_numeric_value": 200.0
        ...     }
        ... }
        >>> m.move_problematic_values_to_misc_param(biosample_problematic)
        {'id': 'test2', 'type': 'nmdc:Biosample', 'misc_param': [{'type': 'nmdc:PropertyAssertion', 'has_attribute_label': 'abs_air_humidity', 'has_attribute_id': 'MIXS:0000122', 'has_numeric_value': 75.0, 'has_unit': 'kPa'}, {'type': 'nmdc:PropertyAssertion', 'has_attribute_label': 'diss_oxygen', 'has_attribute_id': 'MIXS:0000119', 'has_numeric_value': 8.0, 'has_unit': 'mL/L'}, {'type': 'nmdc:PropertyAssertion', 'has_attribute_label': 'solar_irradiance', 'has_attribute_id': 'MIXS:0000112', 'has_numeric_value': 200.0, 'has_unit': 'W/m2'}]}
        """
        problematic_units = (
            ("abs_air_humidity", "MIXS:0000122", "kPa"),
            ("diss_oxygen", "MIXS:0000119", "mL/L"),
            ("solar_irradiance", "MIXS:0000112", "W/m2"),
        )
        for slot_name, slot_uri, bad_unit in problematic_units:
            qv = biosample.get(slot_name)
            if isinstance(qv, dict) and qv.get("type") == "nmdc:QuantityValue":
                if qv.get("has_unit") == bad_unit:
                    # Move to misc_param as PropertyAssertion
                    pa = {
                        "type": "nmdc:PropertyAssertion",
                        "has_attribute_label": slot_name,
                        "has_attribute_id": slot_uri,
                    }
                    if "has_numeric_value" in qv:
                        pa["has_numeric_value"] = qv["has_numeric_value"]
                    if "has_unit" in qv:
                        pa["has_unit"] = qv["has_unit"]
                    if "has_raw_value" in qv:
                        pa["has_raw_value"] = qv["has_raw_value"]

                    # Ensure misc_param is a list and append the new PropertyAssertion
                    if "misc_param" not in biosample or not isinstance(biosample["misc_param"], list):
                        biosample["misc_param"] = []
                    biosample["misc_param"].append(pa)
                    # Remove the problematic QuantityValue from its original slot
                    del biosample[slot_name]
        return biosample



from nmdc_schema.migrators.migrator_base import MigratorBase

class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.8.0.part_4"
    _to_version = "11.8.0.part_5"

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""
        self.adapter.do_for_each_document("material_processing_set", self.validate_sub_sampling_process_slot)

    def validate_sub_sampling_process_slot(self, material_processing_record: dict) -> None:
        r"""
        If the material processing record is of type SubSamplingProcess it should not have a `non_polar_layer` slot.
        If it does, raise a ValueError.
        
        >>> m = Migrator()
        >>> m.validate_sub_sampling_process_slot({"id": 123, "type": "nmdc:SubSamplingProcess"})
        >>> m.validate_sub_sampling_process_slot({"id": 123, "type": "nmdc:SubSamplingProcess", "non_polar_layer": "some_value"})
        Traceback (most recent call last):
        ...
        ValueError: `non_polar_layer` is not an allowed slot for SubSamplingProcess and is present in the material processing record 123
        """
        
        mp_record_id = material_processing_record.get("id")
        # Check if the material processing record is of type SubSamplingProcess
        if material_processing_record.get("type") == "nmdc:SubSamplingProcess":
            # Check if the non_polar_layer slot is present
            if "non_polar_layer" in material_processing_record:
                raise ValueError(f"`non_polar_layer` is not an allowed slot for SubSamplingProcess and is present in the material processing record {mp_record_id}")


import re
from nmdc_schema.migrators.migrator_base import MigratorBase


DOI_URL_PATTERN = r"^https?:\/\/[a-zA-Z\.]+\/10\."


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "7.7.2"
    _to_version = "7.8.0"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document("study_set", [self.replace_doi_field_with_award_dois_list_field])

    def replace_doi_field_with_award_dois_list_field(self, study: dict) -> dict:
        r"""
        Removes the `doi` field. If `doi.has_raw_value` contained a specific type of URL string, this function also adds
        an `award_dois` field whose value is a list containing a CURIE string derived from that URL string.

        >>> m = Migrator()
        >>> m.replace_doi_field_with_award_dois_list_field({'id': 123})  # no `doi` field
        {'id': 123}
        >>> m.replace_doi_field_with_award_dois_list_field({'id': 123, 'doi': {'has_raw_value': 'not-a-url'}})
        {'id': 123}
        >>> m.replace_doi_field_with_award_dois_list_field(
        ...     {'id': 123, 'doi': {'has_raw_value': 'https://example.com/10.other_stuff'}}
        ... )
        {'id': 123, 'award_dois': ['doi:10.other_stuff']}
        """

        self.logger.info(f"Starting migration of {study['id']}")
        if "doi" in study:
            match = re.search(DOI_URL_PATTERN, study["doi"]["has_raw_value"])
            if match:
                start_index = match.end()
                as_curie = f"doi:10.{study['doi']['has_raw_value'][start_index:]}"
                study["award_dois"] = [as_curie]
            del study["doi"]
        return study



from abc import ABC, abstractmethod
from typing import Optional, Callable, List, Any, Union


class AdapterBase(ABC):
    r"""
    Abstract base class containing method stubs related to manipulating a "database" represented by either
    a Python data dictionary or a MongoDB database.
    """

    def __init__(
        self,
        on_collection_created: Optional[Callable[[str], Any]] = None,
        on_collection_renamed: Optional[Callable[[str, str], Any]] = None,
        on_collection_deleted: Optional[Callable[[str], Any]] = None,
    ):
        r"""
        Binds callback functions to the class instance.

        Note: This is a "concrete" method within an "abstract" class.

        :param on_collection_created: Callback function that will be invoked after a collection is created.
                                      It will be passed the name of the collection.
        :param on_collection_renamed: Callback function that will be invoked after a collection is renamed.
                                      It will be passed the original name and new name of the collection.
        :param on_collection_deleted: Callback function that will be invoked after a collection is deleted.
                                      It will be passed the name of the collection.
        """
        self.on_collection_created = on_collection_created
        self.on_collection_renamed = on_collection_renamed
        self.on_collection_deleted = on_collection_deleted

    @abstractmethod
    def create_collection(self, collection_name: str) -> None:
        r"""
        Creates an empty collection having the specified name, if no collection by that name exists.
        Also invokes `self.on_collection_created`, if defined, passing to it the name of the collection.
        """
        pass

    @abstractmethod
    def rename_collection(self, current_name: str, new_name: str) -> None:
        r"""
        Renames the specified collection, if it exists, so that it has the specified new name.
        Also invokes `self.on_collection_renamed`, if defined, passing to it the old and new names of the collection.
        """
        pass

    @abstractmethod
    def delete_collection(self, collection_name: str) -> None:
        r"""
        Deletes the collection having the specified name, if such a collection exists.
        Also invokes `self.on_collection_deleted`, if defined, passing to it the name of the collection.
        """
        pass

    @abstractmethod
    def insert_document(self, collection_name: str, document: dict) -> None:
        r"""
        Inserts the specified document into the collection having the specified name.
        """
        pass

    @abstractmethod
    def get_document_having_value_in_field(
        self, collection_name: str, field_name: str, value: str
    ) -> Optional[dict]:
        r"""
        Retrieves the first document from the specified collection, having the specified value in the specified field.

        Note: This only supports top-level fields (e.g. `_id`, `depth`), not nested fields (e.g. `depth.has_unit`).
        """
        pass

    @abstractmethod
    def get_document_having_one_of_values_in_field(
        self, collection_name: str, field_name: str, values: List[str]
    ) -> Optional[dict]:
        r"""
        Retrieves the first document from the specified collection, having any one of the specified values in the
        specified field.

        Note: This only supports top-level fields (e.g. `_id`, `depth`), not nested fields (e.g. `depth.has_unit`).
        """
        pass

    @abstractmethod
    def delete_documents_having_value_in_field(
        self, collection_name: str, field_name: str, value: str
    ) -> int:
        r"""
        Deletes all documents from the specified collection, having the specified value in the specified field;
        and returns the number of documents that were deleted.

        Note: This only supports top-level fields (e.g. `_id`, `depth`), not nested fields (e.g. `depth.has_unit`).
        """
        pass

    @abstractmethod
    def process_each_document(
        self, collection_name: str, pipeline: List[Callable[[dict], dict]]
    ) -> None:
        r"""
        Passes each document in the specified collection through the specified processing pipeline—in which
        the output of any given function is the input to the function after it—and stores the final output
        back in the collection, replacing the original document.
        """
        pass

    @abstractmethod
    def set_field_of_each_document(
        self, collection_name: str, field_name: str, value: Union[None, str, int, float, bool],
    ) -> None:
        r"""
        Assigns the specified value to the specified field of each document in the collection.
        This method is a specialized alternative to the `process_each_document` method.
        """
        pass

    @abstractmethod
    def remove_field_from_each_document(
        self,
        collection_name: str,
        field_name: str,
    ) -> None:
        r"""
        Removes the specified field from each document in the collection.
        """
        pass

    @abstractmethod
    def do_for_each_document(
        self, collection_name: str, action: Callable[[dict], None]
    ) -> None:
        r"""
        Passes each document in the specified collection to the specified function. This method was designed
        to facilitate iterating over all documents in a collection without actually modifying them.
        """
        pass

    @abstractmethod
    def copy_value_from_field_to_field_in_each_document(
        self,
        collection_name: str,
        source_field_name: str,
        destination_field_name: str,
    ) -> None:
        r"""
        For each document in the collection that has the source field, copy the value of the source field
        into the destination field, creating the destination field if it doesn't already exist.
        """
        pass




from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    """
    Migrates data from X to PR176, namely changes the permissible values of the failure_where slot in the read_qc_analysis_set collection.

    Should be run after migrator_from_10_2_0_to_11_0_0_part_09.py.
    """
    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        # Populate the "collection-to-transformers" map for this specific migration.
        agenda = dict(
            # We can assume that all the documents in the read_qc_analysis_set collection are of the type ReadQcAnalysisActivity, and any OmicsProcessing failures should be mapped to NucleotideSequencing
            read_qc_analysis_set=[self.standardize_failure_where_enum_ns], 
        )

        for collection_name, pipeline in agenda.items():
            self.adapter.process_each_document(collection_name=collection_name, pipeline=pipeline)


    def standardize_failure_where_enum_ns(self, read_qc_analysis: dict):
        r"""
        Transforms the values of the qc_failure_where slot to the new permissible values of the FailureWhereEnum, for sets related to NucleatideSequencing
        
        >>> m = Migrator()
        >>> m.standardize_failure_where_enum_ns({'id': 123, 'has_failure_categorization': [{'qc_failure_where': 'OmicsProcessing'}]})
        {'id': 123, 'has_failure_categorization': [{'qc_failure_where': 'NucleotideSequencing'}]}
        """

        failure_where_translation_dict = {
            "OmicsProcessing": "NucleotideSequencing",
            "Pooling": "Pooling",
            "Extraction": "Extraction",
            "LibraryPreparation": "LibraryPreparation",
            "MetagenomeAssembly": "MetagenomeAssembly",
            "MetatranscriptomeActivity": "MetatranscriptomeExpressionAnalysis",
            "MagsAnalysisActivity": "MagsAnalysis",
            "ReadQcAnalysisActivity": "ReadQcAnalysis",
            "ReadBasedTaxonomyAnalysisActivity": "ReadBasedTaxonomyAnalysis",
            "MetagenomeAnnotationActivity": "MetagenomeAnnotation",
        }

        if 'has_failure_categorization' in read_qc_analysis:
            for failure_categorization in read_qc_analysis['has_failure_categorization']:
                og_value = None  # ensure the variable is defined, since we use it in the error message
                if 'qc_failure_where' in failure_categorization:
                    og_value = failure_categorization.get('qc_failure_where')
                    if og_value in failure_where_translation_dict:
                        failure_categorization['qc_failure_where'] = failure_where_translation_dict.get(og_value)
                else:
                    raise ValueError(f"Invalid 'qc_failure_where' value: {og_value}")
        return read_qc_analysis


from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    # Note: This migrator was introduced via PR 2203 (i.e. https://github.com/microbiomedata/nmdc-schema/pull/2203).
    _from_version = "11.1.0.part_1"
    _to_version = "11.1.0.part_2" 

    def upgrade(self):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        Updates each document in the `functional_annotation_agg` collection so that
        its `metagenome_annotation_id` field is effectively renamed to `was_generated_by`.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> db = {
        ...   'functional_annotation_agg': [
        ...     {'metagenome_annotation_id': 'x', 'other': 'a'},
        ...     {'metagenome_annotation_id': 'y', 'other': 'b'},
        ...   ],
        ... }
        >>> a = DictionaryAdapter(database=db)
        >>> m = Migrator(adapter=a)
        >>> m.upgrade()
        >>> all('metagenome_annotation_id' not in doc for doc in db['functional_annotation_agg'])
        True
        >>> all('was_generated_by' in doc for doc in db['functional_annotation_agg'])
        True
        >>> db['functional_annotation_agg'][0]
        {'other': 'a', 'was_generated_by': 'x'}
        >>> db['functional_annotation_agg'][1]
        {'other': 'b', 'was_generated_by': 'y'}
        """

        self.adapter.copy_value_from_field_to_field_in_each_document(
            collection_name="functional_annotation_agg",
            source_field_name="metagenome_annotation_id",
            destination_field_name="was_generated_by",
        )

        self.adapter.remove_field_from_each_document(
            collection_name="functional_annotation_agg",
            field_name="metagenome_annotation_id",
        )
